# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .const import u4
from minBlepy import floatdtype
import math, numpy as np

signaldtype = np.uint8 # Slightly faster than plain old int.
floatdtype = floatdtype
log2 = math.log(2)

def level5toamp(level):
    return 2 ** ((level - 31) / 4)

def _amptolevel4(amp):
    return max(0, int(round(15 + 2 * math.log(amp) / log2)))

def level4to5(level4):
    return level4 * 2 + 1 # Observe 4-bit 0 is 5-bit 1.

class Shape:

    defaultintrolen = 0
    pyrbotype = dict(buf = [signaldtype], size = u4, introlen = u4)

    @classmethod
    def level4to5(cls, data4, introlen = defaultintrolen):
        return cls(map(level4to5, data4), introlen)

    def __init__(self, g, introlen = defaultintrolen):
        self.buf = np.fromiter(g, signaldtype)
        self.size = self.buf.size
        self.introlen = introlen

    def wavelength(self):
        return self.size - self.introlen

rawtoneshape = 1, 0
toneshape = Shape(rawtoneshape)

def _meansin(x1, x2):
    return (-math.cos(x2) - -math.cos(x1)) / (x2 - x1)

def _sinsliceamp(i, n, skew):
    return (_meansin(*(2 * math.pi * (i + off + skew) / n for off in [-.5, .5])) + 1) / 2

def _sinuslevel4(steps, maxlevel4, skew):
    maxamp = level5toamp(level4to5(maxlevel4))
    for step in range(steps):
        # For each step, the level that's closest to its ideal mean amp:
        yield _amptolevel4(maxamp * _sinsliceamp(step, steps, skew))

level4tosinus5shape = tuple(Shape.level4to5(_sinuslevel4(8, level4, 0)) for level4 in range(16))
level4totone5shape = tuple(Shape.level4to5(level4 * x for x in rawtoneshape) for level4 in range(16))

def makesample5shape(data, signed, is4bit):
    if is4bit:
        data4 = data
    else:
        if signed:
            data = [(d + 0x80) & 0xff for d in data]
        d0 = min(data)
        data4 = (_amptolevel4(((d - d0) + .5) / 0x100) for d in data)
    return Shape.level4to5(data4, len(data) - 1)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from . import BufferFiller
from ..iface import AmpScale, Config, Platform, Stream
from ..nod import Node
from ..out import FloatStream, StereoInfo
from diapyr import types
import logging, outjack.portaudioclient as pac

log = logging.getLogger(__name__)

class PortAudioClient(pac.PortAudioClient, Platform):

    @types(Config, StereoInfo)
    def __init__(self, config, stereoinfo):
        config = config.PortAudio
        super().__init__(stereoinfo.getoutchans.size, config.outputrate, config.buffersize, config.ringsize, config.coupling)

class PortAudioStream(Node, Stream, metaclass = AmpScale):

    log2maxpeaktopeak = 1

    @types(StereoInfo, FloatStream, PortAudioClient)
    def __init__(self, stereoinfo, wavs, client):
        super().__init__()
        self.chancount = stereoinfo.getoutchans.size
        self.wavs = wavs
        self.client = client

    def start(self):
        self.client.activate()
        self.filler = BufferFiller(self.chancount, self.client.buffersize, self.client.current_output_buffer, self.client.send_and_get_output_buffer)

    def callimpl(self):
        self.filler([self.chain(wav) for wav in self.wavs])

    def flush(self):
        pass

    def stop(self):
        self.client.deactivate()

def configure(di):
    di.add(PortAudioClient)
    di.add(PortAudioStream)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Chip, Config, Prerecorded, Roll, Stream, Timer
from .ym2149 import PhysicalRegisters
from diapyr import types
from splut.bg import MainBackground

class Bundle:

    def __init__(self, prerecorded, registers):
        self.prerecorded = prerecorded
        self.registers = registers

    def __iter__(self):
        yield from self.prerecorded.frames(self.registers)

class LogicalBundle(Bundle):

    @types(Prerecorded, [Chip])
    def __init__(self, prerecorded, chips):
        super().__init__(prerecorded, {chip.param: chip.channels for chip in chips})

class PhysicalBundle(Bundle):

    @types(Prerecorded, PhysicalRegisters)
    def __init__(self, prerecorded, registers):
        super().__init__(prerecorded, registers)

class Player(MainBackground):

    @types(Config, Bundle, Roll, Timer, Stream)
    def __init__(self, config, bundle, roll, timer, stream):
        super().__init__(config)
        self.updaterate = config.updaterate
        self.bundle = bundle
        self.roll = roll
        self.timer = timer
        self.stream = stream

    def __call__(self):
        for _ in self.bundle:
            if self.quit:
                break
            self.roll.update()
            for b in self.timer.blocksforperiod(self.updaterate):
                self.stream.call(b)
        self.stream.flush()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
from .nod import BufNode, Node
import logging

log = logging.getLogger(__name__)

class BinMix(BufNode):

    def __init__(self, tone, noise, toneflagreg, noiseflagreg):
        super().__init__(BufType.signal)
        self.tone = tone
        self.noise = noise
        self.toneflagreg = toneflagreg
        self.noiseflagreg = noiseflagreg

    def callimpl(self):
        # The truth table options are {AND, OR, XOR}.
        # Other functions are negations of these, the 2 constants, or not symmetric.
        # XOR sounds just like noise so it can't be that.
        # AND and OR have the same frequency spectrum so either will sound good.
        # We use AND as zero is preferred over envelope, see qbmixenv:
        noiseflag = self.noiseflagreg.value
        if self.toneflagreg.value:
            self.blockbuf.copybuf(self.chain(self.tone))
            if noiseflag:
                self.blockbuf.andbuf(self.chain(self.noise))
        elif noiseflag:
            self.blockbuf.copybuf(self.chain(self.noise))
        else:
            # Fixed and variable levels should work, see qanlgmix and qenvpbuf:
            self.blockbuf.fill_i1(1)

class Multiplexer(Node):

    def __init__(self, buftype, streams):
        super().__init__()
        self.multi = buftype()
        self.channels = len(streams)
        self.streams = streams

    def callimpl(self):
        for i, stream in enumerate(self.streams):
            buf = self.chain(stream)
            if not i:
                # BufNode can't do this because size is not framecount:
                size = len(buf)
                multi = self.multi.ensureandcrop(size * self.channels)
            multi.putstrided(i, i + self.channels * size, self.channels, buf.buf)
        return multi

class IdealMixer(BufNode):

    def __init__(self, container, log2maxpeaktopeak, chipamps):
        log.debug("Mix is trivial: %s", not chipamps.nontrivial)
        buftype = BufType.float
        if chipamps.nontrivial:
            if len(container) != chipamps.size():
                raise Exception(f"Expected {len(container)} chipamps but got: {chipamps.size()}")
            self.contrib = buftype()
            self.callimpl = self.nontrivialcallimpl
        else:
            self.callimpl = self.trivialcallimpl
        super().__init__(buftype)
        self.datum = buftype.dtype(2 ** (log2maxpeaktopeak - 1.5)) # Half power point, very close to -3 dB.
        self.container = container
        self.chipamps = chipamps

    def nontrivialcallimpl(self):
        self.blockbuf.fill_same(self.datum)
        contrib = self.contrib.ensureandcrop(self.block.framecount)
        for buf, amp in zip(self.chain(self.container), self.chain(self.chipamps)):
            if amp:
                contrib.copybuf(buf)
                contrib.mul(amp)
                self.blockbuf.subbuf(contrib)

    def trivialcallimpl(self):
        self.blockbuf.fill_same(self.datum)
        for buf in self.chain(self.container):
            self.blockbuf.subbuf(buf)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

# cython: language_level=3

cimport numpy as np

cdef extern from "resid/siddefs.h":

    ctypedef unsigned reg8
    ctypedef int cycle_count

cdef extern from "resid/sid.h":

    cdef cppclass SID:
        SID()
        int clock(cycle_count& delta_t, short* buf, int n)
        void write(reg8 offset, reg8 value)

cdef class NativeSID:

    cdef SID sid

    def write(self, reg8 offset, reg8 value):
        self.sid.write(offset, value)

    def clock(self, np.ndarray[short] buf):
        cdef cycle_count delta_t = 0x7fffffff
        return self.sid.clock(delta_t, &buf[0], len(buf))



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .const import u4
from .shapes import floatdtype, signaldtype
from diapyr.util import enum
from pyrbo import generic, LOCAL, turbo, T
import numpy as np

@enum(
    ['float', floatdtype],
    ['int16', np.int16],
    ['short', np.short],
    ['signal', signaldtype],
)
class BufType:

    def __init__(self, _, dtype):
        self.dtype = dtype

    def __call__(self):
        return MasterBuf(self.dtype)

groupsets = {T: [{t.dtype for t in BufType.enum}]}

class Buf(metaclass = generic):

    def _turbo(**types):
        return turbo(types = dict(self = dict(buf = [T]), **types), groupsets = groupsets)

    def __init__(self, buf):
        self.buf = buf

    def __len__(self):
        return len(self.buf)

    def last(self):
        return self.buf[-1]

    @_turbo(endframe = u4, that = dict(buf = [T]), i = u4)
    def copyasprefix(self, endframe, that):
        self_buf = that_buf = LOCAL
        for i in range(endframe):
            self_buf[i] = that_buf[i]

    @_turbo(that = dict(buf = [T]), startframe = u4, endframe = u4, i = u4)
    def copywindow(self, that, startframe, endframe):
        self_buf = that_buf = LOCAL
        for i in range(endframe - startframe):
            self_buf[i] = that_buf[startframe]
            startframe += 1

    @_turbo(startframe = u4, endframe = u4, value = T)
    def fillpart(self, startframe, endframe, value):
        self_buf = LOCAL
        while startframe < endframe:
            self_buf[startframe] = value
            startframe += 1

    @_turbo(startframe = u4, endframe = u4, thatnp = [T], j = u4)
    def partcopyintonp(self, startframe, endframe, thatnp):
        self_buf = LOCAL
        for j in range(endframe - startframe):
            thatnp[j] = self_buf[startframe]
            startframe += 1

    @_turbo(value = np.int8, i = u4, v = T)
    def fill_i1(self, value):
        self_buf = py_self_buf = LOCAL
        v = value # Cast once.
        for i in range(py_self_buf.size):
            self_buf[i] = v

    @_turbo(value = T, i = u4)
    def fill_same(self, value):
        self_buf = py_self_buf = LOCAL
        for i in range(py_self_buf.size):
            self_buf[i] = value

    @_turbo(start = u4, end = u4, step = u4, data = [T], j = u4)
    def putstrided(self, start, end, step, data):
        self_buf = LOCAL
        j = 0
        while start < end:
            self_buf[start] = data[j]
            start += step
            j += 1

    def ceildiv(self, divisor, alreadynegated = False):
        if not alreadynegated:
            self.buf *= -1
        self.buf //= divisor
        self.buf *= -1

    def mulbuf(self, that):
        self.buf *= that.buf

    @_turbo(that = dict(buf = [signaldtype]), lookup = [T], i = u4)
    def mapbuf(self, that, lookup):
        self_buf = that_buf = py_that_buf = LOCAL
        for i in range(py_that_buf.size):
            self_buf[i] = lookup[that_buf[i]]

    def add(self, value):
        self.buf += value

    def mul(self, value):
        self.buf *= value

    def subbuf(self, that):
        self.buf -= that.buf

    def andbuf(self, that):
        self.buf &= that.buf

    def copybuf(self, that):
        self.buf[:] = that.buf

    def tofile(self, fileobj):
        # TODO LATER: Reinstate tofile when fixed upstream, see https://github.com/numpy/numpy/issues/7380 for details.
        fileobj.write(self.buf.tobytes())

    def differentiate(self, lastofprev, that):
        self.copybuf(that)
        self.buf[0] -= lastofprev
        self.buf[1:] -= that.buf[:-1]

    def tolist(self): # For tests.
        return list(self.buf)

class MasterBuf:

    def __init__(self, dtype):
        self.bufcls = Buf[T:dtype]
        self.dtype = dtype
        self.setsize(0)

    def setsize(self, size):
        self.buf = np.empty(size, self.dtype)
        self.bufobj = self.bufcls(self.buf)
        self.size = size

    def ensureandcrop(self, framecount):
        if self.size > framecount:
            return self.bufcls(self.buf[:framecount])
        if self.size < framecount:
            # Ideally we would resize in-place, but that can fall foul of numpy reference counting:
            self.setsize(framecount)
        return self.bufobj



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

# cython: language_level=3

from libc.string cimport memset
from . cimport ctime

cdef extern from "alsa/global.h":
    pass

cdef extern from "alsa/output.h":
    pass

cdef extern from "alsa/input.h":
    pass

cdef extern from "alsa/conf.h":
    pass

cdef extern from "alsa/timer.h":
    pass

cdef extern from "alsa/seq_event.h":

    ctypedef struct snd_seq_ev_note_t:
        unsigned char channel
        unsigned char note
        unsigned char velocity

    ctypedef struct snd_seq_ev_ctrl_t:
        unsigned char channel
        unsigned int param
        signed int value

cdef union snd_seq_event_data:
    snd_seq_ev_note_t note
    snd_seq_ev_ctrl_t control

cdef extern from "alsa/seq_event.h":

    ctypedef unsigned char snd_seq_event_type_t

    ctypedef struct snd_seq_addr_t:
        unsigned char client
        unsigned char port

    ctypedef struct snd_seq_event_t:
        snd_seq_event_type_t type
        unsigned char queue
        snd_seq_addr_t dest
        snd_seq_event_data data

cdef extern from "alsa/seq.h":

    ctypedef struct snd_seq_t:
        pass # Opaque.

    DEF SND_SEQ_OPEN_OUTPUT = 1
    DEF SND_SEQ_OPEN_INPUT = 2
    DEF SND_SEQ_PORT_CAP_WRITE = 1 << 1
    DEF SND_SEQ_PORT_CAP_SUBS_WRITE = 1 << 6 # XXX: What is this?
    DEF SND_SEQ_PORT_TYPE_SYNTHESIZER = 1 << 18

    DEF SND_SEQ_QUEUE_DIRECT = 253

    int snd_seq_open(snd_seq_t**, const char*, int, int)
    int snd_seq_event_input(snd_seq_t*, snd_seq_event_t**) nogil
    int snd_seq_event_output_direct(snd_seq_t*, snd_seq_event_t*)
    int snd_seq_client_id(snd_seq_t*)

cdef extern from "alsa/seqmid.h":

    int snd_seq_set_client_name(snd_seq_t*, const char*)
    int snd_seq_create_simple_port(snd_seq_t*, const char*, unsigned int, unsigned int)

SND_SEQ_EVENT_NOTEON = 6
SND_SEQ_EVENT_NOTEOFF = 7
SND_SEQ_EVENT_CONTROLLER = 10
SND_SEQ_EVENT_PGMCHANGE = 11
SND_SEQ_EVENT_PITCHBEND = 13
SND_SEQ_EVENT_USR0 = 90

cdef class Event:

    cdef readonly double time
    cdef readonly snd_seq_event_type_t type
    cdef readonly unsigned char channel

    cdef initevent(self, ctime.timeval* time, snd_seq_event_type_t type, unsigned char channel):
        self.time = time.tv_sec + time.tv_usec / 1e6
        self.type = type
        self.channel = channel # XXX: Can we also get the client?

cdef class Note(Event):

    cdef readonly unsigned char note
    cdef readonly unsigned char velocity

    cdef init(self, ctime.timeval* time, snd_seq_event_type_t type, snd_seq_ev_note_t* data):
        self.initevent(time, type, data.channel)
        self.note = data.note
        self.velocity = data.velocity

cdef class Ctrl(Event):

    cdef readonly unsigned int param
    cdef readonly signed int value

    cdef init(self, ctime.timeval* time, snd_seq_event_type_t type, snd_seq_ev_ctrl_t* data):
        self.initevent(time, type, data.channel)
        self.param = data.param
        self.value = data.value

cdef class Client:

    cdef snd_seq_t* handle
    cdef int portid

    def __init__(self, const char* client_name, const char* port_name):
        if snd_seq_open(&(self.handle), 'default', SND_SEQ_OPEN_INPUT | SND_SEQ_OPEN_OUTPUT, 0):
            raise Exception('Failed to open ALSA.')
        snd_seq_set_client_name(self.handle, client_name)
        self.portid = snd_seq_create_simple_port(self.handle, port_name, SND_SEQ_PORT_CAP_WRITE | SND_SEQ_PORT_CAP_SUBS_WRITE, SND_SEQ_PORT_TYPE_SYNTHESIZER)

    def event_input(self):
        cdef snd_seq_event_t* event
        cdef ctime.timeval now
        cdef Note note
        cdef Ctrl ctrl
        while True:
            with nogil:
                snd_seq_event_input(self.handle, &event)
                ctime.gettimeofday(&now, NULL)
            if event == NULL: # Renoise triggers this when it is closed.
                continue
            # XXX: Can we turn these ifs into a lookup?
            if SND_SEQ_EVENT_NOTEON == event.type or SND_SEQ_EVENT_NOTEOFF == event.type:
                note = Note.__new__(Note)
                note.init(&now, event.type, <snd_seq_ev_note_t*> &(event.data))
                return note
            elif SND_SEQ_EVENT_CONTROLLER == event.type or SND_SEQ_EVENT_PGMCHANGE == event.type or SND_SEQ_EVENT_PITCHBEND == event.type:
                ctrl = Ctrl.__new__(Ctrl)
                ctrl.init(&now, event.type, <snd_seq_ev_ctrl_t*> &(event.data))
                return ctrl
            elif SND_SEQ_EVENT_USR0 == event.type:
                return

    def interrupt(self):
        cdef snd_seq_event_t event
        memset(&event, 0, sizeof (snd_seq_event_t)) # Compiles to a real sizeof.
        event.type = SND_SEQ_EVENT_USR0
        event.queue = SND_SEQ_QUEUE_DIRECT
        event.dest.client = snd_seq_client_id(self.handle)
        event.dest.port = self.portid
        snd_seq_event_output_direct(self.handle, &event)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .dosound import issleepcommand
import itertools, re

class Line:

    # Does not support quoted whitespace, but we're only interested in numbers:
    pattern = re.compile(r'^(\S+)?(?:\s+(\S+)(?:\s+(\S+))?)?')

    def __init__(self, line):
        self.label, self.directive, self.argstext = self.pattern.search(line).groups()

class Label:

    def __init__(self, bytecode, index):
        self.bytecode = bytecode
        self.index = index

    def __iter__(self):
        return itertools.islice(self.bytecode, self.index, None)

def readlabeltobytecode(f):
    labels = {}
    bytecode = BytecodeBuilder(True)
    for line in map(Line, f):
        if line.label is not None:
            labels[line.label] = bytecode.startingnow() # Last one wins.
        if line.directive is not None:
            bytecode.process(line)
    return labels

class NoSuchLabelException(Exception): pass

def readbytecode(f, findlabel):
    bytecode = None
    for line in map(Line, f):
        if bytecode is None and findlabel == line.label: # XXX: Support terminating colon?
            bytecode = BytecodeBuilder(False) # And fall through to next clause if there is a directive.
        if bytecode is not None and line.directive is not None:
            bytecode.process(line)
            if bytecode.hasterminator():
                break
    if bytecode is None:
        raise NoSuchLabelException(findlabel)
    return bytecode.bytes

class UnsupportedDirectiveException(Exception): pass

class BytecodeBuilder:

    @staticmethod
    def number(s):
        if s[0] == '%':
            return int(s[1:], 2)
        elif s[0] == '$':
            return int(s[1:], 16)
        else:
            return int(s) # XXX: Could it be octal?

    def __init__(self, aligned):
        self.bytes = []
        self.aligned = aligned

    def startingnow(self):
        return Label(self.bytes, len(self.bytes))

    def process(self, line):
        key = line.directive.lower()
        if self.aligned and 'even' == key:
            if len(self.bytes) & 1:
                self.bytes.append(0) # May be used as value by accident.
        elif 'dc.w' == key:
            self.bytes.extend([None, None]) # Unknown endianness.
        elif 'dc.b' == key:
            self.bytes.extend(map(self.number, line.argstext.split(',')))
        else:
            raise UnsupportedDirectiveException(line.directive)

    def hasterminator(self):
        # TODO LATER: Support termination not just at end of line.
        return len(self.bytes) >= 2 and not self.bytes[-1] and issleepcommand(self.bytes[-2])



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

mixer.put() # All off.
A_level.put(0x00)
B_level.put(0x00)
C_level.put(0x00)
setprev(0xff)
A_level.anim(1, 0x0d)
setprev(0xfe)
A_level.anim(2, 0x0d)



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def shapeimpl(self):
    cdef np.uint32_t self_block_framecount = self.block.framecount
    cdef np.ndarray[np.uint8_t] py_self_blockbuf_buf = self.blockbuf.buf
    cdef np.uint8_t* self_blockbuf_buf = &py_self_blockbuf_buf[0]
    cdef np.uint8_t self_eager = self.eager
    cdef np.int32_t self_index = self.index
    cdef np.uint32_t self_periodreg_value = self.periodreg.value
    cdef np.uint32_t self_progress = self.progress
    cdef np.uint32_t self_scale = self.scale
    cdef np.ndarray[np.uint8_t] py_self_shape_buf = self.shape.buf
    cdef np.uint8_t* self_shape_buf = &py_self_shape_buf[0]
    cdef np.uint32_t self_shape_introlen = self.shape.introlen
    cdef np.uint32_t self_shape_size = self.shape.size
    cdef np.uint32_t self_stepsize = self.stepsize
    cdef np.uint32_t i
    cdef np.uint8_t val
    cdef np.uint32_t j
    cdef np.uint32_t n
    if self_eager:
        self_stepsize = self_periodreg_value * self_scale
    i = 0
    if self_progress < self_stepsize:
        val = self_shape_buf[self_index]
        j = min(self_stepsize - self_progress, self_block_framecount)
        while i < j:
            self_blockbuf_buf[i] = val
            i += 1
    if i == self_block_framecount:
        self_progress += self_block_framecount
    else:
        if not self_eager:
            self_stepsize = self_periodreg_value * self_scale
        n = (self_block_framecount - i) // self_stepsize
        while n:
            self_index += 1
            if self_index == self_shape_size:
                self_index = self_shape_introlen
            val = self_shape_buf[self_index]
            j = i + self_stepsize
            while i < j:
                self_blockbuf_buf[i] = val
                i += 1
            n -= 1
        if i == self_block_framecount:
            self_progress = self_stepsize
        else:
            self_index += 1
            if self_index == self_shape_size:
                self_index = self_shape_introlen
            val = self_shape_buf[self_index]
            self_progress = self_block_framecount - i
            while i < self_block_framecount:
                self_blockbuf_buf[i] = val
                i += 1
    return self_index, self_progress, self_stepsize



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def rtoneimpl(self, np.uint32_t prescaler, np.uint32_t etdr):
    cdef np.uint32_t self_block_framecount = self.block.framecount
    cdef np.ndarray[np.uint8_t] py_self_blockbuf_buf = self.blockbuf.buf
    cdef np.uint8_t* self_blockbuf_buf = &py_self_blockbuf_buf[0]
    cdef np.uint64_t self_chipimplclock = self.chipimplclock
    cdef np.int32_t self_index = self.index
    cdef np.int32_t self_maincounter = self.maincounter
    cdef np.int64_t self_mfpclock = self.mfpclock
    cdef np.uint32_t self_precounterxmfp = self.precounterxmfp
    cdef np.uint32_t self_repeat = self.repeat
    cdef np.ndarray[np.uint8_t] py_self_shape_buf = self.shape.buf
    cdef np.uint8_t* self_shape_buf = &py_self_shape_buf[0]
    cdef np.uint32_t self_shape_introlen = self.shape.introlen
    cdef np.uint32_t self_shape_size = self.shape.size
    cdef np.uint64_t chunksizexmfp
    cdef np.uint64_t stepsizexmfp
    cdef np.int64_t nextstepxmfp
    cdef np.uint32_t i
    cdef np.int64_t numerator
    cdef np.uint32_t j
    cdef np.uint8_t val
    chunksizexmfp = self_chipimplclock * prescaler
    stepsizexmfp = chunksizexmfp * etdr
    nextstepxmfp = chunksizexmfp * self_maincounter + self_precounterxmfp - chunksizexmfp
    i = 0
    while True:
        numerator = nextstepxmfp + self_mfpclock - 1
        if numerator >= 0:
            j = min(numerator // self_mfpclock, self_block_framecount)
            val = self_shape_buf[self_index]
            while i < j:
                self_blockbuf_buf[i] = val
                i += 1
            if j == self_block_framecount:
                break
        nextstepxmfp += stepsizexmfp
        self_index += 1
        if self_index == self_shape_size:
            self_repeat += 1
            self_index = self_shape_introlen
    nextstepxmfp -= self_mfpclock * self_block_framecount
    self_maincounter = 1
    while nextstepxmfp < 0:
        nextstepxmfp += chunksizexmfp
        self_maincounter -= 1
    self_maincounter += nextstepxmfp // chunksizexmfp
    return self_repeat, self_index, self_maincounter, nextstepxmfp % chunksizexmfp



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from . import BufferFiller
from ..const import clientname
from ..iface import AmpScale, Config, Platform, Stream
from ..nod import Node
from ..out import FloatStream, StereoInfo
from diapyr import types
import outjack.jackclient as jc, logging

log = logging.getLogger(__name__)

class JackClient(jc.JackClient, Platform):

    @types(Config, StereoInfo)
    def __init__(self, config, stereoinfo):
        portcount = stereoinfo.getoutchans.size + config.SID.enabled # FIXME: This is a hack.
        super().__init__(clientname, portcount, config.jackringsize, config.jackcoupling)

    def start(self):
        super().start()
        log.debug(
            "JACK block size: %s or %.3f seconds",
            self.buffersize,
            self.buffersize / self.outputrate)

class JackStream(Stream, Node, metaclass = AmpScale):

    log2maxpeaktopeak = 1

    @types(Config, [FloatStream], JackClient)
    def __init__(self, config, streams, client):
        super().__init__()
        self.systemchannelcount = config.systemchannelcount
        for stream in streams:
            for chanindex in range(stream.chancount):
                client.port_register_output(f"{stream.streamname}_{1 + chanindex}")
        self.streams = streams
        self.client = client

    def start(self):
        self.client.activate()
        for stream in self.streams:
            # Connect all system channels, cycling over our streams if necessary:
            for syschanindex in range(self.systemchannelcount):
                chanindex = syschanindex % stream.chancount
                self.client.connect(f"{clientname}:{stream.streamname}_{1 + chanindex}", f"system:playback_{1 + syschanindex}")
        self.filler = BufferFiller(sum(s.chancount for s in self.streams), self.client.buffersize, self.client.current_output_buffer, self.client.send_and_get_output_buffer)

    def callimpl(self):
        self.filler([self.chain(wav) for stream in self.streams for wav in stream])

    def flush(self):
        pass # Nothing to be done.

    def stop(self):
        self.client.deactivate()

def configure(di):
    di.add(JackClient)
    di.add(JackStream)



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def putstrided_float32(self, np.uint32_t start, np.uint32_t end, np.uint32_t step, np.ndarray[np.float32_t] py_data):
    cdef np.ndarray[np.float32_t] py_self_buf = self.buf
    cdef np.float32_t* self_buf = &py_self_buf[0]
    cdef np.float32_t* data = &py_data[0]
    cdef np.uint32_t j
    j = 0
    while start < end:
        self_buf[start] = data[j]
        start += step
        j += 1

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def putstrided_int16(self, np.uint32_t start, np.uint32_t end, np.uint32_t step, np.ndarray[np.int16_t] py_data):
    cdef np.ndarray[np.int16_t] py_self_buf = self.buf
    cdef np.int16_t* self_buf = &py_self_buf[0]
    cdef np.int16_t* data = &py_data[0]
    cdef np.uint32_t j
    j = 0
    while start < end:
        self_buf[start] = data[j]
        start += step
        j += 1

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def putstrided_uint8(self, np.uint32_t start, np.uint32_t end, np.uint32_t step, np.ndarray[np.uint8_t] py_data):
    cdef np.ndarray[np.uint8_t] py_self_buf = self.buf
    cdef np.uint8_t* self_buf = &py_self_buf[0]
    cdef np.uint8_t* data = &py_data[0]
    cdef np.uint32_t j
    j = 0
    while start < end:
        self_buf[start] = data[j]
        start += step
        j += 1



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Prerecorded
import logging, math

log = logging.getLogger(__name__)
refreshrate = 50 # Authentic period.

class BadCommandException(Exception): pass

class Bytecode(Prerecorded):

    pianorollheight = refreshrate

    def __init__(self, bytes, extraseconds):
        self.bytes = bytes
        self.extraseconds = extraseconds

    def frames(self, chip):
        yield from _dosoundimpl(self.bytes, chip)
        if self.extraseconds:
            n = math.ceil(self.extraseconds * refreshrate)
            log.info("Streaming %.3f extra seconds.", n / refreshrate)
            yield from range(n)

def _dosoundimpl(bytecode, chip):
    def g():
        for b in bytecode:
            yield b & 0xff # It's supposed to be bytecode.
    g = g()
    while True:
        ctrl = next(g)
        if ctrl <= 0xF:
            chip.R[ctrl].value = next(g)
        elif 0x80 == ctrl:
            softreg = next(g)
        elif 0x81 == ctrl:
            targetreg = chip.R[next(g)]
            adjust = next(g)
            if adjust >= 0x80:
                adjust -= 0x100 # Convert back to signed.
            last = next(g)
            while True:
                softreg += adjust # Yes, this is done up-front.
                # The real thing simply uses the truncation on overflow:
                targetreg.value = softreg
                yield
                # That's right, if we skip past it we loop forever:
                if last == softreg:
                    break
        elif issleepcommand(ctrl):
            ticks = next(g)
            if not ticks:
                break
            ticks += 1 # Apparently!
            for _ in range(ticks):
                yield
        else:
            raise BadCommandException(ctrl)

def issleepcommand(ctrl):
    return ctrl >= 0x82



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
from .clock import ClockInfo
from .iface import AmpScale, Chip, Config, Tuning
from .lurlene import convenient
from .minblep import MinBleps
from .native.resid import NativeSID
from .nod import Node
from .out import FloatStream
from .reg import Reg, regproperty
from diapyr import types
from lurlene import topitch
from minBlepy.minblep import Translator

PAL = 4.43361875e6 * 4 / 18
NTSC = 3.579545e6 * 4 / 14

class ChanProxy:

    degree = regproperty(lambda self: self.degreereg)
    pulsewidth = regproperty(lambda self: self.pulsewidthreg)
    pulse = regproperty(lambda self: self.pulsereg)
    triangle = regproperty(lambda self: self.trianglereg)
    gate = regproperty(lambda self: self.gatereg)
    attack = regproperty(lambda self: self.attackreg)
    decay = regproperty(lambda self: self.decayreg)
    sustain = regproperty(lambda self: self.sustainreg)
    release = regproperty(lambda self: self.releasereg)
    filt = regproperty(lambda self: self.filtreg)

    def __init__(self, sidregs, chan, fclk, tuning):
        self.degreereg = Reg()
        pitchreg = Reg().link(topitch, self.degreereg)
        freqreg = Reg().link(tuning.freq, pitchreg)
        fnreg = Reg().link(lambda fout: max(0, min(0xffff, round(fout * (1 << 24) / fclk))), freqreg)
        sidregs[chan, 0].link(lambda fn: fn & 0xff, fnreg)
        sidregs[chan, 1].link(lambda fn: fn >> 8, fnreg)
        self.pulsewidthreg = Reg()
        pwnreg = Reg().link(lambda pwout: max(0, min(0xfff, round(pwout * 40.95))), self.pulsewidthreg)
        sidregs[chan, 2].link(lambda pwn: pwn & 0xff, pwnreg)
        sidregs[chan, 3].link(lambda pwn: pwn >> 8, pwnreg)
        controlreg = Reg(0)
        sidregs[chan, 4].link(lambda control: control & 0xff, controlreg)
        self.pulsereg = Reg()
        controlreg.mlink(0x40, lambda b: -b, self.pulsereg)
        self.trianglereg = Reg()
        controlreg.mlink(0x10, lambda b: -b, self.trianglereg)
        self.gatereg = Reg()
        controlreg.mlink(0x01, lambda gate: gate, self.gatereg)
        adreg = Reg(0)
        srreg = Reg(0)
        sidregs[chan, 5].link(lambda ad: ad, adreg)
        sidregs[chan, 6].link(lambda sr: sr, srreg)
        self.attackreg = Reg()
        self.decayreg = Reg()
        self.sustainreg = Reg()
        self.releasereg = Reg()
        adreg.mlink(0xf0, lambda attack: attack << 4, self.attackreg)
        adreg.mlink(0x0f, lambda decay: decay, self.decayreg)
        srreg.mlink(0xf0, lambda sustain: max(0, min(0xf, round(10 ** (sustain / 20) * 0xf))) << 4, self.sustainreg)
        srreg.mlink(0x0f, lambda release: release, self.releasereg)
        self.filtreg = Reg()
        sidregs.resfiltreg.mlink(0x01 << chan, lambda b: -b, self.filtreg)

@convenient(ChanProxy)
class ChipProxy:

    volume = regproperty(lambda self: self._sidregs.volumereg)
    lowpass = regproperty(lambda self: self._sidregs.lowpassreg)
    fcdegree = regproperty(lambda self: self._sidregs.fcdegreereg)

    def __init__(self, chan, chanproxies, sidregs):
        self._chanproxies = chanproxies[chan:] + chanproxies[:chan]
        self._sidregs = sidregs

    def __getitem__(self, index):
        return self._chanproxies[index]

class SID(NativeSID):

    chancount = 3
    regcount = chancount * 7 + 4
    log2maxpeaktopeak = 16

    @types()
    def __init__(self):
        super().__init__()

class SIDBuf(Node):

    def __init__(self, clockinfo, ampscale, sid, minbleps):
        super().__init__()
        self.translator = Translator(clockinfo.implclock, minbleps)
        self.shortmaster = BufType.short()
        self.outmaster = BufType.float()
        self.ampscale = 2 ** (ampscale.log2maxpeaktopeak - sid.log2maxpeaktopeak)
        self.sid = sid

    def callimpl(self):
        _, outcount = self.translator.step(self.block.framecount)
        shortbuf = self.shortmaster.ensureandcrop(outcount)
        self.sid.clock(shortbuf.buf)
        outbuf = self.outmaster.ensureandcrop(outcount)
        outbuf.copybuf(shortbuf)
        outbuf.mul(self.ampscale)
        return outbuf

class SIDStream(FloatStream):

    streamname = 'sid'

    @types(ClockInfo, AmpScale, SID, MinBleps)
    def __init__(self, clockinfo, ampscale, sid, minbleps):
        self.append(SIDBuf(clockinfo, ampscale, sid, minbleps))

class SIDRegs:

    class SIDReg:

        idle = True # No downstream links so always idle.

        def __init__(self, sid, index):
            self.sid = sid
            self.index = index

        def set(self, value):
            self.sid.write(self.index, value)

        def link(self, *args):
            Reg.link(self, *args)

    def __init__(self, sid, tuning):
        self.regs = [self.SIDReg(sid, index) for index in range(sid.regcount)]
        self.fcdegreereg = Reg()
        fcpitchreg = Reg().link(topitch, self.fcdegreereg)
        fcfreqreg = Reg().link(tuning.freq, fcpitchreg)
        fcreg = Reg().link(lambda freq: max(0, min(0x7ff, round(freq * .16))), fcfreqreg)
        self.regs[0x15].link(lambda fc: fc, fcreg)
        self.regs[0x16].link(lambda fc: fc >> 3, fcreg)
        self.resfiltreg = Reg(0)
        self.regs[0x17].link(lambda x: x, self.resfiltreg)
        modevolreg = Reg(0)
        self.regs[0x18].link(lambda x: x, modevolreg)
        self.volumereg = Reg()
        modevolreg.mlink(0x0f, lambda volume: max(0, min(0xf, round(volume))), self.volumereg)
        self.lowpassreg = Reg()
        modevolreg.mlink(0x10, lambda b: -b, self.lowpassreg)

    def __getitem__(self, key):
        chan, offset = key
        return self.regs[chan * 7 + offset]

class SIDChip(Chip):

    param = 'sid'

    @types(Config, Tuning, SID)
    def __init__(self, config, tuning, sid):
        fclk = config.SID['clock']
        sidregs = SIDRegs(sid, tuning)
        chanproxies = [ChanProxy(sidregs, chan, fclk, tuning) for chan in range(sid.chancount)]
        self.channels = [ChipProxy(chan, chanproxies, sidregs) for chan in range(sid.chancount)]

def configure(di):
    di.add(SID)
    di.add(SIDStream)
    di.add(SIDChip)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

class BufferFiller:

    def __init__(self, portcount, buffersize, init, flip):
        self.portcount = portcount
        self.buffersize = buffersize
        self._newbuf(init)
        self.flip = flip

    def __call__(self, outbufs):
        n = len(outbufs[0])
        i = 0
        while i < n:
            m = min(n - i, self.buffersize - self.cursor)
            for portindex in range(self.portcount):
                self.outbuf[portindex, self.cursor:self.cursor + m] = outbufs[portindex].buf[i:i + m]
            self.cursor += m
            i += m
            if self.cursor == self.buffersize:
                self._newbuf(self.flip)

    def _newbuf(self, factory):
        outbuf = factory().view()
        outbuf.shape = self.portcount, self.buffersize
        self.outbuf = outbuf
        self.cursor = 0



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import ClockInfo
from .dac import Dac, Level
from .iface import AmpScale, Config
from .lfsr import Lfsr
from .mfp import mfpclock, MFPTimer
from .mix import BinMix
from .nod import Container
from .osc2 import EnvOsc, NoiseOsc, RToneOsc, Shape, ToneOsc
from .reg import Reg, VersionReg
from diapyr import types
import logging

log = logging.getLogger(__name__)
ym2149nzdegrees = 17, 14
# TP, NP and EP are suitable for plugging into the formulas in the datasheet:
TP = lambda f, r: ((r & 0x0f) << 8) | f
NP = lambda p: p & 0x1f
EP = lambda f, r: (r << 8) | f
getlevelmode = lambda l: bool(l & 0x10)

class MixerFlag:

    def __init__(self, bit):
        self.mask = 0x01 << bit

    def __call__(self, m):
        return not (m & self.mask)

class LogicalRegisters:

    @types(Config, ClockInfo)
    def __init__(self, config, clockinfo, minnoiseperiod = 1):
        channels = range(config.chipchannels)
        # TODO: Make effective tone period user-readable without exposing mintoneperiod.
        # Clamping 0 to 1 is authentic for all 3 kinds of period, see qtonpzer, qnoispec, qenvpzer respectively:
        self.toneperiods = [Reg(maxval = config.maxtoneperiod, minval = clockinfo.mintoneperiod) for _ in channels]
        self.noiseperiod = Reg(maxval = config.maxnoiseperiod, minval = minnoiseperiod)
        self.toneflags = [Reg() for _ in channels]
        self.noiseflags = [Reg() for _ in channels]
        self.fixedlevels = [Reg(maxval = 0x0f, minval = 0) for _ in channels]
        self.levelmodes = [Reg() for _ in channels]
        self.envshape = VersionReg(maxval = 0x0f, minval = 0)
        self.envperiod = Reg(maxval = config.maxenvperiod, minval = 1)
        initialmixerflag = MixerFlag(0)(0) # Result is the same whatever the bit.
        for c in channels:
            self.toneperiods[c].value = TP(0, 0)
            self.toneflags[c].value = initialmixerflag
            self.noiseflags[c].value = initialmixerflag
            self.fixedlevels[c].value = 0
            self.levelmodes[c].value = getlevelmode(0)
        self.noiseperiod.value = 0
        self.envperiod.value = EP(0, 0)
        self.envshape.value = 0
        self.timers = tuple(MFPTimer() for _ in channels)

class PhysicalRegisters:

    supportedchannels = 3
    timers = property(lambda self: self.logical.timers)
    levelbase = 0x8

    @types(Config, LogicalRegisters)
    def __init__(self, config, logical):
        # XXX: Add reverse wiring?
        # Like the real thing we have 16 registers, this impl ignores the last 2:
        self.R = tuple(Reg() for _ in range(16)) # Assume all incoming values are in [0, 255].
        # We only have registers for the authentic number of channels:
        for c in range(min(self.supportedchannels, config.chipchannels)):
            logical.toneperiods[c].link(TP, self.R[c * 2], self.R[c * 2 + 1])
            logical.toneflags[c].link(MixerFlag(c), self.R[0x7])
            logical.noiseflags[c].link(MixerFlag(self.supportedchannels + c), self.R[0x7])
            logical.fixedlevels[c].link(lambda l: l & 0x0f, self.R[self.levelbase + c])
            logical.levelmodes[c].link(getlevelmode, self.R[self.levelbase + c])
        logical.noiseperiod.link(NP, self.R[0x6])
        logical.envperiod.link(EP, self.R[0xB], self.R[0xC])
        logical.envshape.link(lambda s: s & 0x0f, self.R[0xD])
        for r in self.R:
            r.value = 0
        self.logical = logical

class YM2149(Container):

    noiseshape = Shape(Lfsr(ym2149nzdegrees))

    @types(Config, ClockInfo, AmpScale, LogicalRegisters)
    def __init__(self, config, clockinfo, ampscale, logical):
        self.scale = clockinfo.scale
        channels = config.chipchannels
        self.oscpause = config.oscpause
        self.clock = clockinfo.implclock
        # Chip-wide signals:
        noise = NoiseOsc(self.scale, logical.noiseperiod, self.noiseshape)
        env = EnvOsc(self.scale, logical.envperiod, logical.envshape)
        # Digital channels from binary to level in [0, 31]:
        tones = [ToneOsc(self.scale, logical.toneperiods[c]) for c in range(channels)]
        rtones = [RToneOsc(mfpclock, self.clock, logical.timers[c], logical.fixedlevels[c]) for c in range(channels)]
        # XXX: Add rtones to maskables?
        self.maskables = tones + [noise, env] # Maskable by mixer and level mode.
        binchans = [BinMix(tones[c], noise, logical.toneflags[c], logical.noiseflags[c]) for c in range(channels)]
        levels = [Level(logical.levelmodes[c], logical.fixedlevels[c], env, binchans[c], rtones[c], logical.timers[c].effect) for c in range(channels)]
        super().__init__([Dac(level, ampscale.log2maxpeaktopeak, channels) for level in levels])

    def callimpl(self):
        result = super().callimpl()
        if not self.oscpause:
            # Pass the block to any nodes that were masked:
            for maskable in self.maskables:
                maskable(self.block, True) # The masked flag tells the node we don't care about output.
        return result



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .dac import NullEffect
from .reg import Reg

prescalers = {1 + i: v for i, v in enumerate([4, 10, 16, 50, 64, 100, 200])}
mfpclock = 2457600

class MFPTimer:

    def __init__(self):
        self.control = Reg()
        self.data = Reg()
        # TODO LATER: Verify that TDR 0 indeed behaves like 0x100.
        self.effectivedata = Reg().link(lambda tdr: tdr if tdr else 0x100, self.data)
        self.control.value = 0
        self.data.value = 0
        self.effect = Reg(NullEffect)
        self.control_data = Reg()
        self.control.link(lambda cd: cd[0], self.control_data)
        self.data.link(lambda cd: cd[1], self.control_data)
        self.freq = Reg()
        # XXX: Should change of wavelength trigger this link?
        self.control_data.link(self._findtcrtdr, self.freq)
        self.prescalerornone = Reg().link(lambda tcr: prescalers.get(tcr), self.control)
        self.repeat = Reg().link(lambda _: None, self.effect)

    def update(self, tcr, tdr, effect):
        self.control_data.value = tcr, tdr
        self.effect.value = effect

    def _findtcrtdr(self, freq):
        if not freq:
            return 0, self.data.value # Stop timer.
        diff = float('inf')
        for tcr, prescaler in prescalers.items():
            prescaler *= self.effect.value.wavelength # Avoid having to multiply twice.
            etdr = int(round(mfpclock / (freq * prescaler)))
            if 1 <= etdr and etdr <= 0x100:
                d = abs(mfpclock / (etdr * prescaler) - freq)
                if d < diff:
                    tcrtdr = tcr, etdr & 0xff
                    diff = d
        return tcrtdr

    def _getnormperiod(self):
        return prescalers[self.control.value] * self.effectivedata.value * self.effect.value.wavelength

    def getfreq(self): # Currently only called when effect is not None.
        return mfpclock / self._getnormperiod() if self.control.value else 0



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

'Compile Dosound DSL scripts to bytecode for playback on a real Atari.'
from functools import reduce
import operator, sys

class Reg:

    def __init__(self, data, index, xform):
        self.data = data
        self.index = index
        self.xform = xform

    def put(self, *value):
        self.data.bytecode.extend([self.index, self.xform(*value)])

    def anim(self, preadjust, last):
        self.data.bytecode.extend([0x81, self.index, preadjust & 0xff, last])
        # If prev equals last we do a full cycle rather than nothing, see dosound0:
        value = self.data.prev
        while True:
            value = (value + preadjust) & 0xff
            self.data.totalticks += 1
            if value == last or value == self.data.prev:
                break

class UnsupportedTicksException(Exception): pass

class Data:

    def __init__(self):
        self.index = 0
        self.totalticks = 0
        self.bytecode = []

    def reg(self, xform = lambda x: x):
        r = Reg(self, self.index, xform)
        self.index += 1
        return r

    def setprev(self, prev):
        self.bytecode.extend([0x80, prev])
        self.prev = prev

    def sleep(self, ticks):
        if ticks < 2:
            raise UnsupportedTicksException(ticks)
        while ticks:
            part = min(256, ticks)
            self.bytecode.extend([0x82, part - 1])
            self.totalticks += part
            ticks -= part

    def save(self, f):
        w = lambda v: f.write(bytes(v))
        w([self.totalticks >> 8, self.totalticks & 0xff])
        w(self.bytecode)
        w([0x82, 0])

class Globals:

    def __init__(g, data):
        g.A_fine, g.A_rough, g.B_fine, g.B_rough, g.C_fine, g.C_rough, g.N_period = (data.reg() for _ in range(7))
        g.mixer = data.reg(lambda *v: 0x3f & ~reduce(operator.or_, v, 0))
        g.A_level, g.B_level, g.C_level, g.E_fine, g.E_rough, g.E_shape = (data.reg() for _ in range(6))
        g.A_tone, g.B_tone, g.C_tone, g.A_noise, g.B_noise, g.C_noise = (0x01 << i for i in range(6))
        g.setprev = data.setprev
        g.sleep = data.sleep

def main():
    for inpath in sys.argv[1:]:
        outpath = inpath[:inpath.rindex('.')] + '.dsd'
        print(outpath, file=sys.stderr)
        data = Data()
        exec(compile(open(inpath).read(), inpath, 'exec'), Globals(data).__dict__)
        with open(outpath, 'wb') as f:
            data.save(f)

if '__main__' == __name__:
    main()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import stclock
from .dac import DigiDrumEffect, NullEffect, PWMEffect, SinusEffect
from .iface import Config, YMFile
from .lha import unlha
from .shapes import makesample5shape
from .ym2149 import PhysicalRegisters
from contextlib import contextmanager
from diapyr import types
from foyndation import singleton
from functools import partial
from lagoon.util import onerror
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
import logging, os, struct, sys

log = logging.getLogger(__name__)

class LoopInfo:

    def __init__(self, frame, offset):
        self.frame = frame
        self.offset = offset

class YMFileException(Exception): pass

class YM:

    checkstr = b'LeOnArD!'
    wordstruct = struct.Struct('>H')
    lwordstruct = struct.Struct('>I')
    lwordlestruct = struct.Struct('<I')

    @staticmethod
    def logignoringloopinfo():
        log.debug('Ignoring loop info.')

    def __init__(self, f, expectcheckstr):
        log.debug("Format ID: %s", self.formatid)
        if expectcheckstr:
            if self.checkstr != f.read(len(self.checkstr)):
                raise YMFileException('Bad check string.')
        self.frameindex = 0
        self.f = f

    def number(self, struct):
        return struct.unpack(self.f.read(struct.size))[0]

    def word(self):
        return self.number(self.wordstruct)

    def lword(self):
        return self.number(self.lwordstruct)

    def readloopframe(self):
        loopframe = self.lword()
        if loopframe < self.framecount:
            return loopframe
        self.skip(-4)
        loopframe = self.number(self.lwordlestruct)
        if loopframe < self.framecount:
            log.info('Loop frame apparently little-endian.')
            return loopframe
        loopframe = 0
        log.warning("Cannot interpret loop frame, defaulting to %s.", loopframe)
        return loopframe

    def skip(self, n):
        self.f.seek(n, 1)

    def ntstring(self):
        start = self.f.tell()
        while ord(self.f.read(1)):
            pass
        textlen = self.f.tell() - 1 - start
        self.f.seek(start)
        text = self.f.read(textlen).decode()
        self.skip(1)
        return text

    def interleavedframe(self):
        frame = [None] * self.framesize
        for i in range(self.framesize - 1):
            frame[i] = ord(self.f.read(1))
            self.skip(-1 + self.framecount)
        frame[self.framesize - 1] = ord(self.f.read(1))
        self.skip(-(self.framesize - 1) * self.framecount)
        return frame

    def simpleframe(self):
        return [c for c in self.f.read(self.framesize)]

    def step(self):
        frame = self.frameobj(self)
        self.frameindex += 1
        return frame

    def __iter__(self):
        while self.frameindex < self.framecount:
            yield self.step()
        if self.loopinfo is None:
            return
        while True:
            if not (self.frameindex - self.framecount) % (self.framecount - self.loopinfo.frame):
                log.debug("Looping to frame %s.", self.loopinfo.frame)
                self.f.seek(self.loopinfo.offset)
            yield self.step()

    def close(self):
        self.f.close()

class PlainFrame:

    def __init__(self, ym):
        self.data = ym.readframe()

    def applydata(self, chip):
        for i, x in enumerate(self.data):
            if 0xD != i or 255 != x:
                chip.R[i].value = x

class YM23(YM):

    framesize = 14
    clock = stclock
    framefreq = 50
    info = ()
    readframe = YM.interleavedframe
    loopinfo = None # Default, overridden in YM3b.
    frameobj = PlainFrame

    def __init__(self, f):
        super().__init__(f, False)
        self.framecount = (os.fstat(f.fileno()).st_size - len(self.formatid)) // self.framesize

class YM2(YM23): # FIXME LATER: Work out format from ST-Sound source, it's not this simple.

    formatid = 'YM2!'

    def __init__(self, f, once):
        super().__init__(f)

class YM3(YM23):

    formatid = 'YM3!'

    def __init__(self, f, once):
        super().__init__(f)

class YM3b(YM23):

    formatid = 'YM3b'

    def __init__(self, f, once):
        super().__init__(f)
        if once:
            self.logignoringloopinfo()
        else:
            self.skip(self.framecount * self.framesize)
            loopframe = self.readloopframe()
            self.skip(-(self.framecount * self.framesize + 4))
            self.loopinfo = LoopInfo(loopframe, self.f.tell() + loopframe)

@singleton
class EternalTTL:

    def decr(self):
        pass

    def exhausted(self):
        pass

class FramesTTL:

    def __init__(self, frames):
        self.frames = frames

    def decr(self):
        self.frames -= 1

    def exhausted(self):
        return not self.frames

class SampleTTL:

    def __init__(self, repeatreg):
        self.repeatreg = repeatreg

    def decr(self):
        pass

    def exhausted(self):
        # XXX: Somehow ensure last sample value gets a full timer period?
        return self.repeatreg.value

class YM56(YM):

    framesize = 16

    def __init__(self, f, once):
        super().__init__(f, True)
        self.framecount = self.lword()
        interleaved, ddsigned, dd4bit = (songattr & (1 << b) for songattr in [self.lword()] for b in range(3))
        samplecount = self.word()
        self.clock = self.lword()
        self.framefreq = self.word()
        loopframe = self.readloopframe()
        self.skip(self.word()) # Future expansion.
        log.debug("Read %s samples.", samplecount)
        self.samples = [makesample5shape(self.f.read(self.lword()), ddsigned, dd4bit) for _ in range(samplecount)]
        self.info = tuple(self.ntstring() for _ in range(3))
        dataoffset = self.f.tell()
        self.readframe = self.interleavedframe if interleaved else self.simpleframe
        if once:
            self.logignoringloopinfo()
            self.loopinfo = None
        elif interleaved:
            self.loopinfo = LoopInfo(loopframe, dataoffset + loopframe)
        else:
            self.loopinfo = LoopInfo(loopframe, dataoffset + loopframe * self.framesize)
        self.timerttls = [EternalTTL] * PhysicalRegisters.supportedchannels

    @contextmanager
    def processtimerttls(self, chip):
        for ttl in self.timerttls:
            ttl.decr()
        yield self.timerttls
        for chan, timer in enumerate(chip.timers):
            if self.timerttls[chan].exhausted():
                timer.effect.value = NullEffect
                self.timerttls[chan] = EternalTTL

class Frame56(PlainFrame):

    def __init__(self, ym):
        super().__init__(ym)
        self.index = ym.frameindex
        self.flags = ym

    def applydata(self, chip):
        with self.flags.processtimerttls(chip) as timerttls:
            for chan, tcr, tdr, effect, ttl in self.updatetimers(chip):
                chip.timers[chan].update(tcr, tdr, effect)
                timerttls[chan] = ttl
        super().applydata(chip)

    def sampleornone(self, chan):
        index = self.data[0x8 + chan] & 0x1f
        try:
            return self.flags.samples[index]
        except IndexError:
            log.warning("Bad sample index: %s", index)

class Frame5(Frame56):

    def updatetimers(self, chip):
        if self.data[0x1] & 0x30:
            chan = ((self.data[0x1] & 0x30) >> 4) - 1
            tcr = (self.data[0x6] & 0xe0) >> 5
            tdr = self.data[0xE]
            yield chan, tcr, tdr, PWMEffect, FramesTTL(1)
        if self.data[0x3] & 0x30:
            chan = ((self.data[0x3] & 0x30) >> 4) - 1
            sample = self.sampleornone(chan)
            if sample is not None:
                tcr = (self.data[0x8] & 0xe0) >> 5
                tdr = self.data[0xF]
                yield chan, tcr, tdr, DigiDrumEffect(sample), SampleTTL(chip.timers[chan].repeat)

class Frame6(Frame56):

    def updatetimers(self, chip):
        for r, rr, rrr in [0x1, 0x6, 0xE], [0x3, 0x8, 0xF]:
            if self.data[r] & 0x30:
                chan = ((self.data[r] & 0x30) >> 4) - 1
                fx = self.data[r] & 0xc0
                if 0x00 == fx:
                    tcr = (self.data[rr] & 0xe0) >> 5
                    tdr = self.data[rrr]
                    yield chan, tcr, tdr, PWMEffect, FramesTTL(1)
                if 0x40 == fx:
                    sample = self.sampleornone(chan)
                    if sample is not None:
                        tcr = (self.data[rr] & 0xe0) >> 5
                        tdr = self.data[rrr]
                        yield chan, tcr, tdr, DigiDrumEffect(sample), SampleTTL(chip.timers[chan].repeat)
                if 0x80 == fx:
                    tcr = (self.data[rr] & 0xe0) >> 5
                    tdr = self.data[rrr]
                    yield chan, tcr, tdr, SinusEffect, FramesTTL(1)
                if self.flags.logsyncbuzzer and 0xc0 == fx:
                    log.warning("Sync-buzzer at frame %s.", self.index)
                    self.flags.logsyncbuzzer = False

class YM5(YM56):

    formatid = 'YM5!'
    frameobj = Frame5

class YM6(YM56):

    formatid = 'YM6!'
    frameobj = Frame6
    logsyncbuzzer = True

class YMOpen(YMFile):

    impls = {i.formatid.encode(): i for i in [YM2, YM3, YM3b, YM5, YM6]}
    magic = b'YM'

    @types(Config)
    def __init__(self, config):
        self.path = Path(config.inpath)
        self.once = config.ignoreloop
        self.unpackedfactory = partial(UnpackedFile, _builtin_lha if config.builtin_lha else _real_lha)

    def start(self):
        self.startimpl()
        for info in self.ym.info:
            log.info(info)
        self.nominalclock = self.ym.clock
        self.pianorollheight = self.updaterate = self.ym.framefreq

    def startimpl(self):
        self.f = self.path.open('rb')
        with onerror(self.f.close):
            if self.magic == self.f.read(2):
                self.ym = self.impls[self.magic + self.f.read(2)](self.f, self.once)
                return
        self.f.close()
        self.f = self.unpackedfactory(self.path)
        with onerror(self.f.close):
            self.ym = self.impls[self.f.read(4)](self.f, self.once)

    def frames(self, chip):
        for frame in self.ym:
            frame.applydata(chip)
            yield

    def stop(self):
        self.f.close()

class UnpackedFile:

    def __init__(self, lhaimpl, path):
        self.tmpdir = Path(mkdtemp())
        with onerror(self._clean):
            lhaimpl(path, self.tmpdir)
            p, = self.tmpdir.iterdir()
            self.f = p.open('rb')

    def _clean(self):
        log.debug("Deleting temporary folder: %s", self.tmpdir)
        rmtree(self.tmpdir)

    def __getattr__(self, name):
        return getattr(self.f, name)

    def close(self):
        self.f.close()
        self._clean()

def _builtin_lha(path, dirpath):
    (dirpath / 'unlha').write_bytes(unlha(path.read_bytes()))

def _real_lha(path, dirpath):
    from lagoon.text import lha
    # Observe we redirect stdout so it doesn't get played:
    lha.x(os.path.abspath(path), cwd = dirpath, stdout = sys.stderr)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .. import minblep, pitch
from ..budgie import readbytecode
from ..clock import ClockInfo
from ..dosound import Bytecode
from ..iface import Config
from ..lurlene import YM2149Chip
from ..out import StereoInfo, YMStream
from ..util import MainThread
from ..vis import NullRoll, RollImpl
from ..ym2149 import LogicalRegisters, PhysicalRegisters, YM2149
from diapyr import DI, types
from lurlene.context import Context

@types(Config, this = Bytecode)
def srcbytecodefactory(config):
    with open(config.inpath) as f:
        return Bytecode(readbytecode(f, config.srclabel), config.dosoundextraseconds)

def boot(configname):
    di = DI()
    di.add(configname.loadconfig)
    di.add(di)
    di.add(ClockInfo)
    di.add(StereoInfo)
    di.add(YM2149)
    di.add(PhysicalRegisters)
    di.add(LogicalRegisters)
    di.add(minblep.loadorcreate)
    di.add(YMStream)
    pitch.configure(di)
    di.add(Context)
    config = di(Config)
    di.add(RollImpl if config.pianorollenabled else NullRoll)
    di.add(MainThread)
    di.add(YM2149Chip)
    if config.SID.enabled:
        from . import resid
        resid.configure(di)
    return config, di



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import ClockInfo
from .iface import Platform
from .minblep import MinBleps
from .nod import Block
from .timer import MinBlockRateTimer, SimpleTimer
from diapyr import types

class SyncTimer(SimpleTimer):
    'Fill platform buffer most efficiently.'

    @types(Platform, MinBleps, ClockInfo)
    def __init__(self, platform, minbleps, clockinfo):
        self.naiverate = clockinfo.implclock
        super().__init__(self.naiverate)
        self.buffersize = platform.buffersize
        self.naivex = 0
        self.bufferx = 0
        self.minbleps = minbleps

    def blocksforperiod(self, refreshrate):
        wholeperiodblock, = super().blocksforperiod(refreshrate)
        naiveN = wholeperiodblock.framecount
        while naiveN:
            naiven = min(naiveN, self.minbleps.getminnaiven(self.naivex, self.buffersize - self.bufferx))
            yield Block(naiven)
            self.bufferx = (self.bufferx + self.minbleps.getoutcount(self.naivex, naiven)) % self.buffersize
            self.naivex = (self.naivex + naiven) % self.naiverate
            naiveN -= naiven

class ChipTimer(MinBlockRateTimer):

    @types(ClockInfo)
    def __init__(self, clockinfo):
        super().__init__(clockinfo.implclock, 100)

class SimpleChipTimer(SimpleTimer):
    'One block per update.'

    @types(ClockInfo)
    def __init__(self, clockinfo):
        super().__init__(clockinfo.implclock)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
import errno, struct, sys

class Wave16:

    bytespersample = 2
    hugefilesize = 0x80000000
    buftype = BufType.int16
    formats = {2: '<H', 4: '<I'}

    def __init__(self, path, rate, channels):
        if '-' == path:
            self.f = sys.stdout.buffer
        else:
            self.f = open(path, 'wb') # Binary.
        self.f.write(b'RIFF')
        self.riffsizeoff = 4
        self.writeriffsize(self.hugefilesize)
        self.f.write(b'WAVEfmt ') # Observe trailing space.
        self.writen(16) # Chunk data size.
        self.writen(1, 2) # PCM (uncompressed).
        self.writen(channels, 2)
        self.writen(rate)
        bytesperframe = self.bytespersample * channels
        self.writen(rate * bytesperframe) # Bytes per second.
        self.writen(bytesperframe, 2)
        self.writen(self.bytespersample * 8, 2) # Bits per sample.
        self.f.write(b'data')
        self.datasizeoff = 40
        self.writedatasize(self.hugefilesize)
        self.adjustsizes()

    def writeriffsize(self, filesize):
        self.writen(filesize - (self.riffsizeoff + 4))

    def writedatasize(self, filesize):
        self.writen(filesize - (self.datasizeoff + 4))

    def writen(self, n, size = 4):
        self.f.write(struct.pack(self.formats[size], n))

    def block(self, buf):
        buf.tofile(self.f)
        self.adjustsizes()

    def adjustsizes(self):
        try:
            filesize = self.f.tell()
            if not filesize:
                return # It's probably /dev/null so give up.
        except IOError as e:
            if errno.ESPIPE != e.errno:
                raise
            return # Leave huge sizes.
        self.f.seek(self.riffsizeoff)
        self.writeriffsize(filesize)
        self.f.seek(self.datasizeoff)
        self.writedatasize(filesize)
        self.f.seek(filesize)

    def flush(self):
        self.f.flush()

    def close(self):
        self.f.close()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config
from argparse import ArgumentParser
from aridity import NoSuchPathException
from aridity.config import ConfigCtrl
from aridity.model import Boolean, Resource, wrap
from diapyr import DI, types, UnsatisfiableRequestException
from io import StringIO
import logging, sys

log = logging.getLogger(__name__)

class ConfigName:

    def __init__(self, *params, args = sys.argv[1:], name = 'root'):
        parser = ArgumentParser()
        parser.add_argument('--config', action = 'append', default = [])
        parser.add_argument('--ignore-settings', action = 'store_true')
        for param in params:
            parser.add_argument(param)
        self.additems = parser.parse_args(args)
        self.resource = Resource(__name__, f"{name}.arid")

    @types(DI, this = Config)
    def loadconfig(self, di):
        cc = ConfigCtrl()
        config = cc._loadappconfig('pym2149', self.resource)
        config.diref = DIRef(di)
        if not self.additems.ignore_settings:
            try:
                cc.loadsettings()
            except FileNotFoundError as e:
                log.warning("Settings not found: %s", e)
        for name, value in self.additems.__dict__.items():
            if 'config' == name:
                for text in value:
                    (-config).load(StringIO(text))
            else:
                setattr(config, name, value)
        return config

class DIRef:

    def __init__(self, di):
        self.di = di

    def __call__(self, scope, resolvable):
        try:
            return wrap(self.di(resolvable.resolve(scope).scalar))
        except UnsatisfiableRequestException:
            raise NoSuchPathException

def isvalue(scope, resolvable):
    try:
        resolvable.resolve(scope)
        val = True
    except NoSuchPathException:
        val = False
    return Boolean(val)

def pyattr(scope, objresolvable, attrresolvable):
    return wrap(getattr(objresolvable.resolve(scope).scalar, attrresolvable.resolve(scope).textvalue))

def py(scope, coderesolvable):
    return wrap(eval(coderesolvable.resolve(scope).textvalue, {}))



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import ClockInfo
from .dac import NullEffect
from .iface import Chip, Config, Prerecorded, Tuning
from .reg import Reg, regproperty
from .ym2149 import LogicalRegisters
from diapyr import types
from diapyr.start import Started
from foyndation import innerclass
from lurlene import topitch
from lurlene.bridge import LiveCodingBridge
from lurlene.context import Context
import logging

log = logging.getLogger(__name__)

def _convenience(name):
    def fget(self):
        return getattr(self._chanproxies[0], name)
    def fset(self, value):
        setattr(self._chanproxies[0], name, value)
    return property(fget, fset)

def convenient(sourcecls):
    def f(targetcls):
        for name in dir(sourcecls):
            if '_' != name[0]:
                setattr(targetcls, name, _convenience(name))
        return targetcls
    return f

class ChanProxy:

    tonefreq = regproperty(lambda self: self.tonefreqreg)
    level = regproperty(lambda self: self.levelreg)
    noiseflag = regproperty(lambda self: self._chip.noiseflags[self._chan])
    toneflag = regproperty(lambda self: self._chip.toneflags[self._chan])
    toneperiod = regproperty(lambda self: self.toneperiodreg)
    tonepitch = regproperty(lambda self: self.tonepitchreg)
    tonedegree = regproperty(lambda self: self.tonedegreereg)
    envflag = regproperty(lambda self: self._chip.levelmodes[self._chan])
    timerfreq = regproperty(lambda self: self._chip.timers[self._chan].freq)
    timereffect = regproperty(lambda self: self._chip.timers[self._chan].effect)

    def __init__(self, chip, chan, clock, tuning):
        self.tonedegreereg = Reg()
        self.tonepitchreg = Reg().link(topitch, self.tonedegreereg)
        self.tonefreqreg = Reg().link(tuning.freq, self.tonepitchreg)
        self.toneperiodreg = Reg().link(clock.toneperiod, self.tonefreqreg)
        chip.toneperiods[chan].link(round, self.toneperiodreg)
        self.levelreg = Reg()
        chip.fixedlevels[chan].link(round, self.levelreg)
        self._chip = chip
        self._chan = chan

    def blank(self):
        self.level = 0
        self.noiseflag = False
        self.toneflag = False
        self.envflag = False
        self.timereffect = NullEffect

class ChipRegs:

    def __init__(self, chip, clock, tuning):
        self.noisefreqreg = Reg()
        self.noiseperiodreg = Reg().link(clock.noiseperiod, self.noisefreqreg)
        chip.noiseperiod.link(round, self.noiseperiodreg)
        self.envdegreereg = Reg()
        self.envpitchreg = Reg().link(topitch, self.envdegreereg)
        self.envfreqreg = Reg().link(tuning.freq, self.envpitchreg)
        self.envperiodreg = Reg().link(lambda f, s: clock.envperiod(f, s), self.envfreqreg, chip.envshape)
        chip.envperiod.link(round, self.envperiodreg)

@convenient(ChanProxy)
class ChipProxy:

    noiseperiod = regproperty(lambda self: self._chipregs.noiseperiodreg)
    envshape = regproperty(lambda self: self._chip.envshape)
    envperiod = regproperty(lambda self: self._chipregs.envperiodreg)
    envpitch = regproperty(lambda self: self._chipregs.envpitchreg)
    envdegree = regproperty(lambda self: self._chipregs.envdegreereg)
    noisefreq = regproperty(lambda self: self._chipregs.noisefreqreg)
    envfreq = regproperty(lambda self: self._chipregs.envfreqreg)

    @innerclass # FIXME: This exposes things that aren't effective.
    class Effective:

        @property
        def envfreq(self):
            return self._clock.envfreq(self._chip.envperiod.value, self._chip.envshape.value)

        @property
        def tonefreq(self):
            return self._clock.tonefreq(self._chip.toneperiods[self._chan].value)

    def __init__(self, chip, chan, chanproxies, chipregs, clock):
        self.effective = self.Effective()
        self._chanproxies = chanproxies[chan:] + chanproxies[:chan]
        self._chip = chip
        self._chipregs = chipregs
        self._clock = clock
        self._chan = chan

    def __getitem__(self, index):
        return self._chanproxies[index]

    def noisepriority(self):
        return not any(chan.noiseflag for chan in self._chanproxies[1:])

class YM2149Chip(Chip):

    param = 'ym'

    @types(Config, LogicalRegisters, ClockInfo, Tuning)
    def __init__(self, config, chip, clock, tuning):
        chans = range(config.chipchannels)
        chipregs = ChipRegs(chip, clock, tuning)
        chanproxies = [ChanProxy(chip, chan, clock, tuning) for chan in chans]
        self.channels = [ChipProxy(chip, chan, chanproxies, chipregs, clock) for chan in chans]

class LurleneBridge(LiveCodingBridge, Prerecorded): pass

@types(Config, Context, this = Started)
def loadcontext(config, context):
    with open(config.inpath) as f:
        context.update(f.read())
    context.flip()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

octet = 8

class Stream:

    masks = [(1 << m) - 1 for m in range(octet + 1)] # Element 0 unused.
    remaining = octet

    def __init__(self, data, cursor = 0, step = 1):
        self.data = data
        self.cursor = cursor
        self.step = step

    def read(self, n):
        x = 0
        while n:
            m = min(n, self.remaining)
            n -= m
            self.remaining -= m
            x = x << m | self.data[self.cursor] >> self.remaining & self.masks[m]
            if not self.remaining:
                self.cursor += self.step
                self.remaining = octet
        return x

    def readcodelen(self):
        x = self.read(3)
        if 7 == x:
            while self.read(1):
                x += 1
        return x

class Tree:

    trivialkey = 1 # The empty code prefixed with a 1.

    @classmethod
    def canonical(cls, codelens):
        lookup = {}
        key = cls.trivialkey
        currentlen = 0
        for l, value in sorted((l, x) for x, l in enumerate(codelens) if l):
            key <<= l - currentlen
            currentlen = l
            lookup[key] = value
            key += 1
        return cls(lookup)

    @classmethod
    def degenerate(cls, value):
        return cls({cls.trivialkey: value})

    def __init__(self, lookup):
        self.lookup = lookup

    def readvalue(self, stream):
        k = self.trivialkey
        while k not in self.lookup:
            k = k << 1 | stream.read(1)
        return self.lookup[k]

class UnsupportedFormatException(Exception): pass

class Interpreter(Tree):

    @classmethod
    def readtree(cls, stream, prefix = 3):
        size = stream.read(5)
        if not size:
            return cls.degenerate(stream.read(5))
        if prefix < size:
            lens = [stream.readcodelen() for _ in range(prefix)]
            zeros = stream.read(2) # XXX: When prefix == size should this be done redundantly?
            lens.extend(0 for _ in range(zeros))
            lens.extend(stream.readcodelen() for _ in range(size - prefix - zeros))
        else:
            lens = [stream.readcodelen() for _ in range(size)]
        return cls.canonical(lens)

    def mainlens(self, stream, lookup = {
        0: lambda s: 1,
        1: lambda s: 3 + s.read(4),
        2: lambda s: 20 + s.read(9),
    }):
        instruction = self.readvalue(stream)
        readzeros = lookup.get(instruction)
        if readzeros is None:
            yield instruction - 2
        else:
            for _ in range(readzeros(stream)):
                yield 0

class Main(Tree):

    @classmethod
    def readtree(cls, stream):
        interpreter = Interpreter.readtree(stream)
        size = stream.read(9)
        if not size:
            return cls.degenerate(stream.read(9))
        lens = []
        while len(lens) < size:
            lens.extend(interpreter.mainlens(stream))
        return cls.canonical(lens)

    def invoke(self, pair, offset):
        x = self.readvalue(pair.stream)
        if x < 256:
            yield x
        else:
            yield from offset.copy(pair, x - 253)

class Offset(Tree):

    @classmethod
    def readtree(cls, stream):
        size = stream.read(4)
        if not size:
            return cls.degenerate(stream.read(4))
        return cls.canonical(stream.readcodelen() for _ in range(size))

    def copy(self, pair, n):
        off = len(pair.buffer) - 1
        x = self.readvalue(pair.stream)
        if x < 2:
            off -= x
        else:
            x -= 1
            off -= (1 << x) | pair.stream.read(x)
        for i in range(off, off + n):
            yield pair.buffer[i]

class Pair:

    def __init__(self, stream):
        self.buffer = []
        self.stream = stream

    def pipe(self, targetlen):
        while len(self.buffer) != targetlen:
            commands = self.stream.read(16)
            main = Main.readtree(self.stream)
            offset = Offset.readtree(self.stream)
            for _ in range(commands):
                self.buffer.extend(main.invoke(self, offset))

def unlha(u):
    if b'-lh5-' != u[2:7] or u[20]:
        raise UnsupportedFormatException
    encstart = 2 + u[0]
    afterenc = encstart + Stream(u, 10, -1).read(32)
    if afterenc != len(u) - 1 or u[afterenc]:
        raise UnsupportedFormatException
    pair = Pair(Stream(u, encstart))
    pair.pipe(Stream(u, 14, -1).read(32))
    return bytes(pair.buffer)



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def partcopyintonp_float32(self, np.uint32_t startframe, np.uint32_t endframe, np.ndarray[np.float32_t] py_thatnp):
    cdef np.ndarray[np.float32_t] py_self_buf = self.buf
    cdef np.float32_t* self_buf = &py_self_buf[0]
    cdef np.float32_t* thatnp = &py_thatnp[0]
    cdef np.uint32_t j
    for j in range(endframe - startframe):
        thatnp[j] = self_buf[startframe]
        startframe += 1

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def partcopyintonp_int16(self, np.uint32_t startframe, np.uint32_t endframe, np.ndarray[np.int16_t] py_thatnp):
    cdef np.ndarray[np.int16_t] py_self_buf = self.buf
    cdef np.int16_t* self_buf = &py_self_buf[0]
    cdef np.int16_t* thatnp = &py_thatnp[0]
    cdef np.uint32_t j
    for j in range(endframe - startframe):
        thatnp[j] = self_buf[startframe]
        startframe += 1

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def partcopyintonp_uint8(self, np.uint32_t startframe, np.uint32_t endframe, np.ndarray[np.uint8_t] py_thatnp):
    cdef np.ndarray[np.uint8_t] py_self_buf = self.buf
    cdef np.uint8_t* self_buf = &py_self_buf[0]
    cdef np.uint8_t* thatnp = &py_thatnp[0]
    cdef np.uint32_t j
    for j in range(endframe - startframe):
        thatnp[j] = self_buf[startframe]
        startframe += 1



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

'Render a YM file to logging.'
from . import boot
from .. import txt
from ..config import ConfigName
from ..timerimpl import ChipTimer
from ..util import MainThread
from ..ymformat import YMOpen
from ..ymplayer import PhysicalBundle, Player
from diapyr.start import Started
from foyndation import initlogging

def main():
    initlogging()
    config, di = boot(ConfigName('inpath', name = 'txt'))
    with di:
        di.add(YMOpen)
        txt.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

if '__main__' == __name__:
    main()



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def copyasprefix_float32(self, np.uint32_t endframe, that):
    cdef np.ndarray[np.float32_t] py_self_buf = self.buf
    cdef np.float32_t* self_buf = &py_self_buf[0]
    cdef np.ndarray[np.float32_t] py_that_buf = that.buf
    cdef np.float32_t* that_buf = &py_that_buf[0]
    cdef np.uint32_t i
    for i in range(endframe):
        self_buf[i] = that_buf[i]

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def copyasprefix_int16(self, np.uint32_t endframe, that):
    cdef np.ndarray[np.int16_t] py_self_buf = self.buf
    cdef np.int16_t* self_buf = &py_self_buf[0]
    cdef np.ndarray[np.int16_t] py_that_buf = that.buf
    cdef np.int16_t* that_buf = &py_that_buf[0]
    cdef np.uint32_t i
    for i in range(endframe):
        self_buf[i] = that_buf[i]

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def copyasprefix_uint8(self, np.uint32_t endframe, that):
    cdef np.ndarray[np.uint8_t] py_self_buf = self.buf
    cdef np.uint8_t* self_buf = &py_self_buf[0]
    cdef np.ndarray[np.uint8_t] py_that_buf = that.buf
    cdef np.uint8_t* that_buf = &py_that_buf[0]
    cdef np.uint32_t i
    for i in range(endframe):
        self_buf[i] = that_buf[i]



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
from .clock import ClockInfo
from .iface import AmpScale, Config, Multiplexed, Platform, Stream
from .minblep import MinBleps
from .mix import IdealMixer, Multiplexer
from .nod import Node
from .shapes import floatdtype
from .wav import Wave16
from .ym2149 import YM2149
from diapyr import types
from foyndation import singleton
from minBlepy.minblep import Translator
import logging, numpy as np

log = logging.getLogger(__name__)

class StereoInfo:

    @types(Config)
    def __init__(self, config):
        self.n = config.chipchannels
        if config.stereo:
            self.panlaw = config.panlaw
            self.maxpan = float(config.maxpan)
            self.getoutchans = self.getstaticoutchans
        else:
            self.getoutchans = self.gettrivialoutchans

    def pantoamp(self, outchan, pan):
        return ((1 + (outchan * 2 - 1) * pan) / 2) ** (self.panlaw / 6)

    def staticamp(self, outchan, chipchan):
        return self.pantoamp(outchan, (chipchan * 2 - (self.n - 1)) / (self.n - 1) * self.maxpan)

    def gettrivialoutchans(self):
        return [TrivialOutChannel]
    gettrivialoutchans.size = 1

    def getstaticoutchans(self):
        return [StaticOutChannel([self.staticamp(oc, cc) for cc in range(self.n)]) for oc in range(2)]
    getstaticoutchans.size = 2

class WavWriter(Stream, Node, metaclass = AmpScale):

    log2maxpeaktopeak = 16

    @types(Config, Multiplexed, StereoInfo, Platform)
    def __init__(self, config, wav, stereoinfo, platform):
        super().__init__()
        fclass = Wave16
        self.open = lambda: fclass(config.outpath, platform.outputrate, stereoinfo.getoutchans.size)
        self.roundmaster = BufType.float()
        self.wavmaster = fclass.buftype()
        self.wav = wav

    def start(self):
        self.f = self.open()

    def callimpl(self):
        outbuf = self.chain(self.wav)
        roundbuf = self.roundmaster.ensureandcrop(len(outbuf))
        wavbuf = self.wavmaster.ensureandcrop(len(outbuf))
        np.around(outbuf.buf, out = roundbuf.buf)
        wavbuf.copybuf(roundbuf)
        self.f.block(wavbuf)

    def flush(self):
        self.f.flush()

    def stop(self):
        self.f.close()

class StaticOutChannel(Node):

    nontrivial = True

    def __init__(self, chipamps):
        super().__init__()
        self.chipamps = chipamps

    def size(self):
        return len(self.chipamps)

    def callimpl(self):
        return self.chipamps

@singleton
class TrivialOutChannel:

    nontrivial = False

class FloatStream(list):

    chancount = property(len)

class YMStream(FloatStream):

    streamname = 'ym'

    @types(Config, ClockInfo, YM2149, AmpScale, StereoInfo, MinBleps)
    def __init__(self, config, clockinfo, chip, ampscale, stereoinfo, minbleps):
        naives = [IdealMixer(chip, ampscale.log2maxpeaktopeak, outchan) for outchan in stereoinfo.getoutchans()]
        for naive in naives:
            self.append(WavBuf(clockinfo, naive, minbleps))

class WavBuf(Node):

    @staticmethod
    @types(FloatStream, this = Multiplexed)
    def multi(wavs):
        if 1 == len(wavs):
            wav, = wavs
        else:
            wav = Multiplexer(BufType.float, wavs)
        return wav

    def __init__(self, clockinfo, naive, minbleps):
        super().__init__()
        self.diffmaster = BufType.float()
        self.outmaster = BufType.float()
        # Need space for a whole mixin in case it is rooted at sample outcount:
        self.overflowsize = minbleps.mixinsize
        self.carrybuf = BufType.float().ensureandcrop(self.overflowsize)
        self.translator = Translator(clockinfo.implclock, minbleps)
        self.dc = floatdtype(0) # Last naive value of previous block.
        self.carrybuf.fill_same(self.dc) # Initial carry can be the initial dc level.
        self.naive = naive
        self.minbleps = minbleps

    def callimpl(self):
        # TODO: Unit-test that results do not depend on block size.
        naivebuf = self.chain(self.naive)
        diffbuf = self.diffmaster.ensureandcrop(len(naivebuf))
        diffbuf.differentiate(self.dc, naivebuf)
        naivex, outcount = self.translator.step(self.block.framecount)
        # Make space for all samples we can output plus overflow:
        outsize = outcount + self.overflowsize
        outbuf = self.outmaster.ensureandcrop(outsize)
        # Paste in the carry followed by the carried dc level:
        outbuf.copyasprefix(self.overflowsize, self.carrybuf)
        outbuf.fillpart(self.overflowsize, outsize, self.dc)
        self.minbleps.paste(naivex, diffbuf.buf, outbuf.buf)
        self.carrybuf.copywindow(outbuf, outcount, outsize)
        self.dc = naivebuf.last()
        return self.outmaster.ensureandcrop(outcount)

class WavPlatform(Platform):

    @types(Config)
    def __init__(self, config):
        self.outputrate = config.wavrate

def configure(di):
    di.add(WavPlatform)
    di.add(WavWriter)
    di.add(WavBuf.multi)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
from .const import i4, i8, u1, u4, u8
from .nod import BufNode
from .shapes import Shape, signaldtype, toneshape
from pyrbo import LOCAL, turbo
import itertools, numpy as np

oscnodepyrbotype = dict(
    blockbuf = dict(buf = [signaldtype]),
    block = dict(framecount = u4),
    index = i4,
    shape = Shape.pyrbotype,
)

class ShapeOsc(BufNode):

    progressdtype = u4

    def __init__(self, scale, periodreg):
        super().__init__(BufType.signal)
        self.stepsize = 0 # XXX: Move to reset?
        self.scale = scale
        self.periodreg = periodreg

    def reset(self, shape):
        self.index = -1
        self.progress = np.iinfo(self.progressdtype).max
        self.shape = shape

    def callimpl(self):
        self.index, self.progress, self.stepsize = self.shapeimpl()

    @turbo(
        self = dict(
            oscnodepyrbotype,
            progress = progressdtype,
            stepsize = u4,
            scale = u4,
            periodreg = dict(value = u4),
            eager = u1,
        ),
        i = u4,
        j = u4,
        n = u4,
        val = signaldtype,
    )
    def shapeimpl(self):
        self_blockbuf_buf = self_block_framecount = self_index = self_progress = self_scale = self_periodreg_value = self_shape_buf = self_shape_size = self_shape_introlen = self_eager = LOCAL
        if self_eager:
            self_stepsize = self_periodreg_value * self_scale
        i = 0
        if self_progress < self_stepsize:
            val = self_shape_buf[self_index]
            j = min(self_stepsize - self_progress, self_block_framecount)
            while i < j:
                self_blockbuf_buf[i] = val
                i += 1
        if i == self_block_framecount:
            self_progress += self_block_framecount
        else:
            if not self_eager:
                self_stepsize = self_periodreg_value * self_scale
            n = (self_block_framecount - i) // self_stepsize
            while n:
                self_index += 1
                if self_index == self_shape_size:
                    self_index = self_shape_introlen
                val = self_shape_buf[self_index]
                j = i + self_stepsize
                while i < j:
                    self_blockbuf_buf[i] = val
                    i += 1
                n -= 1
            if i == self_block_framecount:
                self_progress = self_stepsize
            else:
                self_index += 1
                if self_index == self_shape_size:
                    self_index = self_shape_introlen
                val = self_shape_buf[self_index]
                self_progress = self_block_framecount - i
                while i < self_block_framecount:
                    self_blockbuf_buf[i] = val
                    i += 1
        return self_index, self_progress, self_stepsize

class RToneOsc(BufNode):

    effect = None

    def __init__(self, mfpclock, chipimplclock, timer, fixedreg):
        super().__init__(BufType.signal)
        self.mfpclock = mfpclock
        self.chipimplclock = chipimplclock
        self.timer = timer
        self.fixedreg = fixedreg

    def callimpl(self):
        effect = self.timer.effect.value
        if effect is not self.effect:
            # Simulate running interrupt routine synchronously now (start of frame) and subsequently by interrupt:
            self.maincounter = 0
            self.precounterxmfp = None
            self.repeat, self.index = 0, -1 # Reset index in shape.
            self.effect = effect
        self.shape = effect.getshape(self.fixedreg)
        prescalerornone = self.timer.prescalerornone.value
        if prescalerornone is None:
            self.blockbuf.fill_same(self.shape.buf[self.index])
            self.precounterxmfp = None
        else:
            if self.precounterxmfp is None:
                self.precounterxmfp = prescalerornone * self.chipimplclock
            self.repeat, self.index, self.maincounter, self.precounterxmfp = self.rtoneimpl(prescalerornone, self.timer.effectivedata.value)
        self.timer.repeat.value = self.repeat

    @turbo(
        self = dict(
            oscnodepyrbotype,
            mfpclock = i8, # Signed so j doesn't blow up.
            chipimplclock = u8,
            maincounter = i4,
            precounterxmfp = u4,
            repeat = u4,
        ),
        prescaler = u4,
        etdr = u4,
        chunksizexmfp = u8,
        stepsizexmfp = u8,
        nextstepxmfp = i8,
        i = u4,
        j = u4,
        val = signaldtype,
        numerator = i8,
    )
    def rtoneimpl(self, prescaler, etdr):
        self_blockbuf_buf = self_block_framecount = self_mfpclock = self_chipimplclock = self_repeat = self_index = self_maincounter = self_precounterxmfp = self_shape_buf = self_shape_size = self_shape_introlen = LOCAL
        chunksizexmfp = self_chipimplclock * prescaler
        stepsizexmfp = chunksizexmfp * etdr
        nextstepxmfp = chunksizexmfp * self_maincounter + self_precounterxmfp - chunksizexmfp
        i = 0
        while True:
            numerator = nextstepxmfp + self_mfpclock - 1
            if numerator >= 0:
                j = min(numerator // self_mfpclock, self_block_framecount)
                val = self_shape_buf[self_index]
                while i < j:
                    self_blockbuf_buf[i] = val
                    i += 1
                if j == self_block_framecount:
                    break
            nextstepxmfp += stepsizexmfp
            self_index += 1
            if self_index == self_shape_size:
                self_repeat += 1
                self_index = self_shape_introlen
        nextstepxmfp -= self_mfpclock * self_block_framecount
        self_maincounter = 1
        while nextstepxmfp < 0:
            nextstepxmfp += chunksizexmfp
            self_maincounter -= 1
        self_maincounter += nextstepxmfp // chunksizexmfp
        return self_repeat, self_index, self_maincounter, nextstepxmfp % chunksizexmfp

class ToneOsc(ShapeOsc):

    eager = True

    def __init__(self, scale, periodreg):
        scaleofstep = scale * 2 // 2 # Normally half of 16.
        super().__init__(scaleofstep, periodreg)
        self.reset(toneshape)

class NoiseOsc(ShapeOsc):

    eager = False

    def __init__(self, scale, periodreg, shape):
        scaleofstep = scale * 2 # This results in authentic spectrum, see qnoispec.
        super().__init__(scaleofstep, periodreg)
        self.reset(shape)

class EnvOsc(ShapeOsc):

    eager = True
    steps = 32
    shapes = {
        0x0c: Shape(range(steps)),
        0x08: Shape(range(steps - 1, -1, -1)),
        0x0e: Shape(itertools.chain(range(steps), range(steps - 1, -1, -1))),
        0x0a: Shape(itertools.chain(range(steps - 1, -1, -1), range(steps))),
        0x0f: Shape(itertools.chain(range(steps), steps * [0]), steps),
        0x0d: Shape(itertools.chain(range(steps), steps * [steps - 1]), steps),
        0x0b: Shape(itertools.chain(range(steps - 1, -1, -1), steps * [steps - 1]), steps),
        0x09: Shape(itertools.chain(range(steps - 1, -1, -1), steps * [0]), steps),
    }
    for s in range(0x08):
        shapes[s] = shapes[0x0f if s & 0x04 else 0x09]
    del s

    def __init__(self, scale, periodreg, shapereg):
        scaleofstep = scale * 32 // self.steps
        super().__init__(scaleofstep, periodreg)
        self.shapeversion = None
        self.shapereg = shapereg

    def callimpl(self):
        if self.shapeversion != self.shapereg.version:
            self.reset(self.shapes[self.shapereg.value])
            self.shapeversion = self.shapereg.version
        super().callimpl()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

'Render Dosound bytecode to WAV.'
from . import boot
from .. import out
from ..config import ConfigName
from ..dosound import Bytecode
from ..iface import Config
from ..timerimpl import ChipTimer
from ..util import MainThread
from ..ymplayer import PhysicalBundle, Player
from diapyr import types
from diapyr.start import Started
from foyndation import initlogging
import logging

log = logging.getLogger(__name__)

@types(Config, this = Bytecode)
def _dsdbytecodefactory(config):
    with open(config.inpath, 'rb') as f:
        log.debug("Total ticks: %s", (ord(f.read(1)) << 8) | ord(f.read(1)))
        return Bytecode(f.read(), config.dosoundextraseconds)

def main():
    initlogging()
    config, di = boot(ConfigName('inpath', 'outpath', name = 'dsd'))
    with di:
        di.add(_dsdbytecodefactory)
        out.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

if '__main__' == __name__:
    main()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Timer
from .nod import Block
import logging

log = logging.getLogger(__name__)

class SimpleTimer(Timer):

    def __init__(self, clock):
        self.carryticks = 0
        self.clock = clock

    def blocksforperiod(self, refreshrate):
        available = self.carryticks + self.clock
        blockticks = int(round(available / refreshrate))
        self.carryticks = available - blockticks * refreshrate
        yield Block(blockticks)

    def __del__(self):
        if self.carryticks:
            log.warning("Non-zero carry on dispose: %s", self.carryticks)

class MinBlockRateTimer(SimpleTimer):

    def __init__(self, clock, minblockrate):
        # If division not exact, rate will slightly exceed given minimum:
        self.maxblocksize = int(clock // minblockrate)
        if not self.maxblocksize:
            raise Exception(clock, minblockrate)
        super().__init__(clock)

    def blocksforperiod(self, refreshrate):
        wholeperiodblock, = super().blocksforperiod(refreshrate)
        blockticks = wholeperiodblock.framecount
        while blockticks:
            size = min(blockticks, self.maxblocksize)
            b = Block(size)
            blockticks -= size
            yield b



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

class Lfsr:

    def __init__(self, nzdegrees):
        self.mask = sum(1 << (nzd - 1) for nzd in nzdegrees)
        self.x = 1

    def __call__(self):
        bit = self.x & 1
        self.x >>= 1
        if bit:
            self.x ^= self.mask
        return 1 - bit # Authentic, see qnoispec.

    def __iter__(self):
        first = self.x
        while True:
            yield self()
            if first == self.x:
                break



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import AmpScale, Platform, Stream
from diapyr import types

class TxtPlatform(Platform, metaclass = AmpScale):

    # Neither of these is significant for this script:
    outputrate = 44100
    log2maxpeaktopeak = 1

    @types()
    def __init__(self):
        pass

class NullStream(Stream):

    @types()
    def __init__(self):
        pass

    def call(self, block):
        pass

    def flush(self):
        pass

def configure(di):
    di.add(TxtPlatform)
    di.add(NullStream)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

class Link:

    def __init__(self, reg, xform, upstream):
        self.reg = reg
        self.xform = xform
        self.upstream = upstream

    def update(self):
        if self.reg.idle:
            try:
                upstreamvals = [r.value for r in self.upstream]
            except AttributeError:
                return
            self.reg.set(self.xform(*upstreamvals))

class Reg:

    undefined = object()

    def __init__(self, value = undefined, *, maxval = None, minval = None):
        self.links = []
        self.idle = True
        self.maxval = maxval # Friendlier than a mask.
        self.minval = minval # Typically to avoid 0 period.
        if value is not self.undefined:
            self.value = value

    def link(self, xform, *upstream):
        link = Link(self, xform, upstream)
        for r in upstream:
            r.links.append(link)
        return self

    def mlink(self, mask, xform, *upstream):
        negmask = ~mask
        self.link(lambda *args: (negmask & self.value) | (mask & xform(*args)), *upstream)

    def set(self, value):
        if self.maxval is not None:
            value = min(self.maxval, value)
        if self.minval is not None:
            value = max(self.minval, value)
        self._value = value
        self.idle = False
        try:
            for link in self.links:
                link.update()
        finally:
            self.idle = True

    value = property(lambda self: self._value, lambda self, value: self.set(value))

class VersionReg(Reg):

    version = 0

    def set(self, value):
        super().set(value)
        self.version += 1

def regproperty(reg):
    def fget(self):
        return reg(self).value
    def fset(self, value):
        reg(self).value = value
    return property(fget, fset)



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config
from diapyr import types
from splut.bg import MainBackground
import logging, time

log = logging.getLogger(__name__)

class MainThread:

    @types(Config, MainBackground)
    def __init__(self, config, player):
        self.profile = config.profile
        self.trace = config.trace
        self.player = player

    def sleep(self):
        if self.profile or self.trace:
            sleeptime = self.profile.time if self.profile else self.trace
            log.debug("Continue for %.3f seconds.", sleeptime)
            time.sleep(sleeptime)
            log.debug('End of profile, shutting down.')
        else:
            log.debug('Continue until end of data or interrupt.')
            try:
                self.player.future.result()
                log.debug('End of data, shutting down.')
            except KeyboardInterrupt:
                log.debug('Caught interrupt, shutting down.')

class EMA:

    def __init__(self, alpha, initial):
        self.alpha = alpha
        self.value = initial

    def __call__(self, instantaneous):
        self.value = self.alpha * instantaneous + (1 - self.alpha) * self.value
        return self.value

def ceildiv(numerator, denominator):
    return (numerator + denominator - 1) // denominator



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config, Tuning
from diapyr import DI, types
from lurlene.context import Context
import bisect, logging, math

log = logging.getLogger(__name__)

class ContextTuning(Tuning):

    @types(Context, DI)
    def __init__(self, context, di):
        self.context = context
        self.di = di

    def __getattr__(self, name):
        return getattr(self.di(self.context.get('tuning')), name)

    def dispose(self):
        pass

class EqualTemperament:

    @types(Config)
    def __init__(self, config):
        self.reffreq = float(config.referencefrequency)
        self.refmidi = float(config.referencemidinote)

    def freq(self, pitch):
        return self.reffreq * 2 ** ((pitch - self.refmidi) / 12)

    def pitch(self, freq):
        return Pitch(self.refmidi + 12 * math.log(freq / self.reffreq, 2))

class Unequal:

    def __init__(self, refpitch, freqs):
        self.factors = [math.log(g / f, 2) for f, g in zip(freqs, freqs[1:] + [freqs[0] * 2])]
        self.refpitch = refpitch
        self.freqs = freqs

    def freq(self, pitch):
        pitch -= self.refpitch
        pitchindex = math.floor(pitch) % 12
        return self.freqs[pitchindex] * 2 ** (self.factors[pitchindex] * (pitch % 1) + pitch // 12)

    def pitch(self, freq):
        octave = math.floor(math.log(freq / self.freqs[0], 2))
        freq /= 2 ** octave
        pitchindex = bisect.bisect(self.freqs, freq) - 1 # FIXME LATER: Handle out of range.
        x = math.log(freq / self.freqs[pitchindex], 2) / self.factors[pitchindex]
        return Pitch(self.refpitch + octave * 12 + pitchindex + x)

class Meantone(Unequal):
    'Including Pythagorean as a special case when meantonecomma is zero.'

    syntonic = 81 / 80

    @types(Config)
    def __init__(self, config):
        reffreq = float(config.referencefrequency)
        refpitch = float(config.referencemidinote)
        flats = config.meantoneflats
        fifthratio = 1.5 / self.syntonic ** float(config.meantonecomma)
        freqs = [None] * 12
        for fifth in range(-flats, 12 - flats):
            pitch = fifth * 7
            freqs[pitch % 12] = reffreq * fifthratio ** fifth / 2 ** (pitch // 12)
        super().__init__(refpitch, freqs)
        wolfpitch = Pitch(refpitch + 11 - flats)
        log.debug("The fifth %s to %s is a wolf interval.", wolfpitch, wolfpitch + 7)

class FiveLimit(Unequal):
    'Asymmetric variant.'

    @types(Config)
    def __init__(self, config):
        reffreq = float(config.referencefrequency)
        refpitch = float(config.referencemidinote)
        freqs = [None] * 12
        for p5 in range(-1, 2):
            for p3 in range(-1, 3): # Skip first column for asymmetric.
                ratio = 3 ** p3 * 5 ** p5
                ratio /= 2 ** math.floor(math.log(ratio, 2))
                index = round(math.log(ratio, 2) * 12)
                freqs[index] = reffreq * ratio
        super().__init__(refpitch, freqs)

def configure(di):
    di.add(ContextTuning)
    di.add(EqualTemperament)
    di.add(Meantone)
    di.add(FiveLimit)

class Pitch(float):

    names = ('C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B')

    def __add__(self, that):
        return type(self)(float.__add__(self, that))

    def __sub__(self, that):
        return type(self)(float.__sub__(self, that))

    def __str__(self):
        return self.str(10)

    def str(self, mincents):
        nearest = int(math.ceil(self - .5))
        octave = nearest // 12
        note = nearest % 12
        cents = int(round((self - nearest) * 100))
        octave -= 1
        notestr = self.names[note]
        if len(notestr) < 2:
            notestr += '.'
        octavestr = str(octave)
        if len(octavestr) < 2:
            octavestr = '.' + octavestr
        if abs(cents) < mincents:
            centsstr = ' ' * 3
        else:
            centsstr = f"{cents:+3d}"
        return notestr + octavestr + centsstr



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from . import neutral
from lurlene import E

class All:

    shape = 0x0e

    def on(self, ym):
        ym.toneflag = True
        ym.noiseflag = True
        ym.envflag = True
        ym.tonefreq = 1000
        ym.noisefreq = 5000
        if ym.envshape != self.shape:
            ym.envshape = self.shape
        ym.envfreq = 1

sections = [[E(All, '1.5'), neutral, neutral]]
speed = 50



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config, Platform, YMFile
from diapyr import types
import logging

log = logging.getLogger(__name__)
stclock = 2000000
spectrum128crystal = 17734470 # Correct according to service manual.
spectrumclock = spectrum128crystal // 10
defaultscale = 8

class ClockInfo:

    trishapes = frozenset([0xa, 0xe])
    tonescale = 16

    @classmethod
    def _shapescale(cls, shape):
        # Musically, the triangular shapes have twice the scale:
        return 512 if shape in cls.trishapes else 256

    @types(Config, Platform, YMFile)
    def __init__(self, config, platform, ymfile = None):
        self.nomclock = config.nominalclock
        if self.nomclock % config.underclock:
            raise Exception(f"Clock {self.nomclock} not divisible by underclock {config.underclock}.")
        self.implclock = self.nomclock // config.underclock
        if ymfile is not None and self.nomclock != ymfile.nominalclock:
            log.info("Context clock %s overridden to: %s", ymfile.nominalclock, self.nomclock)
        if self.implclock != self.nomclock:
            log.debug("Clock adjusted to %s to take advantage of non-trivial underclock.", self.implclock)
        if config.underclock < 1 or defaultscale % config.underclock:
            raise Exception(f"underclock must be a factor of {defaultscale}.")
        self.scale = defaultscale // config.underclock
        if config.freqclamp:
            # The 0 case just means that 1 is audible:
            self.mintoneperiod = max(1, self._toneperiodclampor0(platform.outputrate))
            log.debug("Minimum tone period: %s", self.mintoneperiod)
        else:
            self.mintoneperiod = 1

    def _toneperiodclampor0(self, outrate):
        # Largest period with frequency strictly greater than Nyquist, or 0 if there isn't one:
        return (self.implclock - 1) // (self.scale * outrate)

    def _convert(self, freqorperiod, scale):
        return self.nomclock / (scale * freqorperiod)

    def toneperiod(self, freq):
        return self._convert(freq, self.tonescale)

    def noiseperiod(self, freq):
        return self._convert(freq, 16) # First notch at freq.

    def envperiod(self, freq, shape):
        return self._convert(freq, self._shapescale(shape))

    def tonefreq(self, period):
        return self._convert(period, self.tonescale)

    def envfreq(self, period, shape):
        return self._convert(period, self._shapescale(shape))



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

import logging

log = logging.getLogger(__name__)

class Block:

    def __init__(self, framecount):
        # If it's a numpy type it can cause overflow in osc module, so ensure arbitrary precision:
        self.framecount = int(framecount)

    def __repr__(self):
        return f"{type(self).__name__}({self.framecount!r})"

class Node:

    def __init__(self):
        self.block = None

    def call(self, block): # I think we don't default masked to False to ensure it is propagated.
        return self(block, False)

    def __call__(self, block, masked):
        if self.block != block:
            self.block = block
            self.masked = masked
            self.result = self.callimpl()
        elif not masked and self.masked:
            log.warning("This node has already executed masked: %s", self)
        return self.result

    def chain(self, node):
        return node(self.block, self.masked)

class Container(Node):

    def __init__(self, nodes):
        super().__init__()
        self.nodes = nodes

    def callimpl(self):
        return [self.chain(node) for node in self.nodes]

    def __len__(self):
        return len(self.nodes)

class BufNode(Node):

    def __init__(self, buftype, channels = 1):
        super().__init__()
        self.masterbuf = buftype()
        self.channels = channels
        self.realcallimpl = self.callimpl
        self.callimpl = self.makeresult

    def makeresult(self):
        self.blockbuf = self.masterbuf.ensureandcrop(self.block.framecount * self.channels)
        resultornone = self.realcallimpl()
        if resultornone is not None:
            return resultornone
        return self.blockbuf



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import ClockInfo
from .dac import DigiDrumEffect, NullEffect, PWMEffect, SinusEffect
from .iface import Config, Roll, Tuning
from .ym2149 import LogicalRegisters
from diapyr import types
from foyndation import singleton

@singleton
class NullRoll(Roll):

    def update(self):
        pass

class RollImpl(Roll):

    line = 0
    shapes = ('\\_',) * 4 + ('/_',) * 4 + ('\\\\', '\\_', '\\/', '\\\u203e', '//', '/\u203e', '/\\', '/_')
    shapeversion = None
    wavestartstr = '~'

    @types(Config, ClockInfo, LogicalRegisters, Tuning)
    def __init__(self, config, clock, chip, tuning):
        self.height = config.pianorollheight
        self.channels = config.chipchannels
        self.periods = config.showperiods # TODO LATER: Command line arg isn't converted to boolean.
        self.stream = config.rollstream
        self.mincents = config.rollmincents
        self.jump = f"\x1b[{self.height}A"
        self.format = ' | '.join(self.channels * ["%7s %1s %2s %1s %2s%1s%7s"])
        self.effects = [None] * self.channels
        self.clock = clock
        self.chip = chip
        self.tuning = tuning

    def _pitchstr(self, freq):
        return self.tuning.pitch(freq).str(self.mincents) if freq else '....   '

    def _effectstartstr(self, c):
        return self.wavestartstr if self.effects[c] is not self.chip.timers[c].effect.value else ''

    def _lhsvals(self, c):
        if self.chip.levelmodes[c].value or self.chip.fixedlevels[c].value:
            tone = self.chip.toneflags[c].value
            noise = self.chip.noiseflags[c].value
            if tone:
                toneperiod = self.chip.toneperiods[c].value
                yield toneperiod if self.periods else self._pitchstr(self.clock.tonefreq(toneperiod))
            else:
                yield ''
            yield '&' if tone and noise else ''
            yield self.chip.noiseperiod.value if noise else ''
            yield '*' if tone or noise else ''
        else:
            yield ''
            yield ''
            yield ''
            yield ''

    def _regularvals(self, c):
        yield from self._lhsvals(c)
        if self.chip.levelmodes[c].value:
            shape = self.chip.envshape.value
            envperiod = self.chip.envperiod.value
            yield self.shapes[shape]
            yield self.wavestartstr if self.shapeversion != self.chip.envshape.version else ''
            yield envperiod if self.periods else self._pitchstr(self.clock.envfreq(envperiod, shape))
        else:
            yield self.chip.fixedlevels[c].value or ''
            yield ''
            yield ''

    def _synthvals(self, c):
        yield from self._lhsvals(c)
        level = self.chip.fixedlevels[c].value
        if level:
            yield level
            yield self._effectstartstr(c)
            yield self._pitchstr(self.chip.timers[c].getfreq())
        else:
            yield ''
            yield '' # XXX: Indicate new effect even if we can't hear it?
            yield ''

    def _ddvals(self, c):
        yield ''
        yield ''
        yield ''
        yield ''
        yield '%%'
        yield self._effectstartstr(c)
        yield ''

    effecttypetovals = {
        DigiDrumEffect: _ddvals,
        type(NullEffect): _regularvals,
        type(PWMEffect): _synthvals,
        type(SinusEffect): _synthvals,
    }

    def _getvals(self, c):
        return self.effecttypetovals[type(self.chip.timers[c].effect.value)](self, c)

    def update(self):
        if self.line == self.height:
            self.stream.write(self.jump)
            self.line = 0
        print(self.format % tuple(v for c in range(self.channels) for v in self._getvals(c)), file = self.stream)
        self.shapeversion = self.chip.envshape.version
        for c in range(self.channels):
            self.effects[c] = self.chip.timers[c].effect.value
        self.line += 1



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

import re

wordpattern = re.compile(r'\S+')
statetobatterypower = {'charging': False, 'fully-charged': False, 'discharging': True}

def batterypower():
    try:
        from lagoon.text import upower
    except ImportError:
        return # Run all tests.
    def states():
        for line in upower.__show_info('/org/freedesktop/UPower/devices/battery_C173').splitlines():
            words = wordpattern.findall(line)
            if 2 == len(words) and 'state:' == words[0]:
                yield words[1]
    states = list(states())
    if states:
        state, = states
        return statetobatterypower[state]



# cython: language_level=3

cimport numpy as np
import cython

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def fill_i1_float32(self, np.int8_t value):
    cdef np.ndarray[np.float32_t] py_self_buf = self.buf
    cdef np.float32_t* self_buf = &py_self_buf[0]
    cdef np.float32_t v
    cdef np.uint32_t i
    v = value # Cast once.
    for i in range(py_self_buf.size):
        self_buf[i] = v

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def fill_i1_int16(self, np.int8_t value):
    cdef np.ndarray[np.int16_t] py_self_buf = self.buf
    cdef np.int16_t* self_buf = &py_self_buf[0]
    cdef np.int16_t v
    cdef np.uint32_t i
    v = value # Cast once.
    for i in range(py_self_buf.size):
        self_buf[i] = v

@cython.boundscheck(False)
@cython.cdivision(True) # Don't check for divide-by-zero.
def fill_i1_uint8(self, np.int8_t value):
    cdef np.ndarray[np.uint8_t] py_self_buf = self.buf
    cdef np.uint8_t* self_buf = &py_self_buf[0]
    cdef np.uint8_t v
    cdef np.uint32_t i
    v = value # Cast once.
    for i in range(py_self_buf.size):
        self_buf[i] = v



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .buf import BufType
from .nod import BufNode
from .shapes import level4to5, level5toamp, level4tosinus5shape, level4totone5shape
from foyndation import singleton
import numpy as np

class Level(BufNode):

    def __init__(self, levelmodereg, fixedreg, env, signal, rtone, timereffectreg):
        super().__init__(BufType.signal) # Must be suitable for use as index downstream.
        self.levelmodereg = levelmodereg
        self.fixedreg = fixedreg
        self.env = env
        self.signal = signal
        self.rtone = rtone
        self.timereffectreg = timereffectreg

    def callimpl(self):
        self.timereffectreg.value.putlevel5(self)

@singleton
class NullEffect:
    'All registers are non-virtual and write directly to chip, the timer does not interfere.'

    def putlevel5(self, node):
        node.blockbuf.copybuf(node.chain(node.signal))
        if node.levelmodereg.value:
            node.blockbuf.mulbuf(node.chain(node.env))
        else:
            # According to block diagram, the level is already 5-bit when combining with binary signal:
            node.blockbuf.mul(level4to5(node.fixedreg.value))

class FixedLevelEffect:
    'Registers levelmodereg and fixedreg are virtual, of which levelmodereg is ignored.'

    def __init__(self):
        self.wavelength, = {shape.wavelength() for shape in self.level4toshape}

    def getshape(self, fixedreg):
        return self.level4toshape[fixedreg.value]

    def putlevel5(self, node):
        node.blockbuf.copybuf(node.chain(node.signal))
        node.blockbuf.mulbuf(node.chain(node.rtone))

@singleton
class PWMEffect(FixedLevelEffect):
    'Interrupt setup/routine alternates fixed level between fixedreg and zero.'

    level4toshape = level4totone5shape

@singleton
class SinusEffect(FixedLevelEffect):
    'Interrupt setup/routine sets fixed level to sample value as scaled by fixedreg.'

    level4toshape = level4tosinus5shape

class DigiDrumEffect:
    'Like SinusEffect but in addition toneflagreg and noiseflagreg are virtual and we ignore them.'

    def __init__(self, sample5shape):
        self.sample5shape = sample5shape

    def getshape(self, fixedreg):
        return self.sample5shape

    def putlevel5(self, node):
        node.blockbuf.copybuf(node.chain(node.rtone))

class Dac(BufNode):

    def __init__(self, level, log2maxpeaktopeak, ampshare):
        buftype = BufType.float
        super().__init__(buftype)
        # We take off .5 so that the peak amplitude is about -3 dB:
        maxpeaktopeak = (2 ** (log2maxpeaktopeak - .5)) / ampshare
        # Lookup of ideal amplitudes:
        self.leveltopeaktopeak = np.fromiter((level5toamp(v) * maxpeaktopeak for v in range(32)), buftype.dtype)
        self.level = level

    def callimpl(self):
        self.blockbuf.mapbuf(self.chain(self.level), self.leveltopeaktopeak)
